"""
Source processing classes
=========================
"""
import datetime
import glob
import hashlib
import json
import logging
import os
import shutil
import time
from collections import namedtuple
from xml.sax import make_parser

import pymongo
from dateutil.parser import parse as parse_datetime
from pymongo import TEXT, ASCENDING
from tqdm import tqdm

from CveXplore.common.config import Configuration
from CveXplore.database.connection.mongo_db import MongoDBConnection
from CveXplore.database.maintenance.Toolkit import generate_title
from CveXplore.database.maintenance.api_handlers import NVDApiHandler
from CveXplore.database.maintenance.content_handlers import CapecHandler, CWEHandler
from CveXplore.database.maintenance.db_action import DatabaseAction
from CveXplore.database.maintenance.file_handlers import XMLFileHandler, JSONFileHandler
from CveXplore.errors.apis import ApiDataRetrievalFailed

date = datetime.datetime.now()
year = date.year + 1

# default config
defaultvalue = {"cwe": "Unknown"}

cveStartYear = Configuration.getCVEStartYear()


class CPEDownloads(NVDApiHandler):
    """
    Class processing CPE source files
    """

    def __init__(self):
        self.feed_type = "CPE"

        super().__init__(self.feed_type)

        self.logger = logging.getLogger("CPEDownloads")

    def file_to_queue(self, *args):
        pass

    @staticmethod
    def process_the_item(item=None):
        if item is None:
            return None

        item = item["cpe"]

        if "cpeName" not in item:
            return None

        title = None

        if "titles" in item:
            for t in item["titles"]:
                if t["lang"] == "en":
                    title = t["title"]

        cpe = {
            "title": title,
            "CveSearchtitle": generate_title(item["cpeName"]),
            "cpe_2_2": item["cpeName"],
            "vendor": item["cpeName"].split(":")[3],
            "product": item["cpeName"].split(":")[4],
            "cpeNameId": item["cpeNameId"],
            "lastModified": parse_datetime(item["lastModified"], ignoretz=True),
            "created": parse_datetime(item["created"], ignoretz=True),
            "deprecated": item["deprecated"],
        }

        version_info = ""
        if "versionStartExcluding" in item:
            cpe["versionStartExcluding"] = item["versionStartExcluding"]
            version_info += cpe["versionStartExcluding"] + "_VSE"
        if "versionStartIncluding" in item:
            cpe["versionStartIncluding"] = item["versionStartIncluding"]
            version_info += cpe["versionStartIncluding"] + "_VSI"
        if "versionEndExcluding" in item:
            cpe["versionEndExcluding"] = item["versionEndExcluding"]
            version_info += cpe["versionEndExcluding"] + "_VEE"
        if "versionEndIncluding" in item:
            cpe["versionEndIncluding"] = item["versionEndIncluding"]
            version_info += cpe["versionEndIncluding"] + "_VEI"

        sha1_hash = hashlib.sha1(
            cpe["cpe_2_2"].encode("utf-8") + version_info.encode("utf-8")
        ).hexdigest()

        cpe["id"] = sha1_hash

        return cpe

    def process_downloads(self, sites=None):
        """
        Method to download and process files
        """

        self.logger.info("Starting download...")

        start_time = time.time()

        self.last_modified = datetime.datetime.now()

        self.logger.debug(
            f"do_process = {self.do_process}; is_update = {self.is_update}"
        )

        if self.do_process:

            if not self.is_update:

                total_results = self.api_handler.get_count(
                    self.api_handler.datasource.CPE
                )

                self.logger.info(f"Preparing to download {total_results} CPE entries")

                with tqdm(
                    desc="Downloading and processing content",
                    total=total_results,
                    position=0,
                    leave=True,
                ) as pbar:
                    for entry in self.api_handler.get_all_data(data_type="cpe"):
                        # do something here with the results...
                        for data_list in tqdm(
                            entry, desc=f"Processing batch", leave=False
                        ):
                            if not isinstance(data_list, ApiDataRetrievalFailed):
                                processed_items = [
                                    self.process_item(item)
                                    for item in data_list["products"]
                                ]
                                self._db_bulk_writer(processed_items)
                                pbar.update(len(data_list["products"]))
                            else:
                                self.logger.error(
                                    f"Retrieval of api data on url: {data_list.args[0]} failed...."
                                )
            else:
                last_mod_start_date = self.database[self.feed_type.lower()].find_one(
                    {}, {"lastModified": 1}, sort=[("lastModified", -1)]
                )

                if "lastModified" in last_mod_start_date:
                    last_mod_start_date = last_mod_start_date[
                        "lastModified"
                    ] + datetime.timedelta(
                        0, 1
                    )  # add one second to prevent false results...
                else:
                    raise KeyError(
                        "Missing field 'lastModified' from database query..."
                    )

                # Get datetime from runtime
                last_mod_end_date = datetime.datetime.now()

                total_results = self.api_handler.get_count(
                    self.api_handler.datasource.CPE,
                    last_mod_start_date=last_mod_start_date,
                    last_mod_end_date=last_mod_end_date,
                )

                self.logger.info(f"Preparing to download {total_results} CPE entries")

                with tqdm(
                    desc="Downloading and processing content",
                    total=total_results,
                    position=0,
                    leave=True,
                ) as pbar:
                    for entry in self.api_handler.get_all_data(
                        data_type="cpe",
                        last_mod_start_date=last_mod_start_date,
                        last_mod_end_date=last_mod_end_date,
                    ):
                        # do something here with the results...
                        for data_list in tqdm(
                            entry, desc=f"Processing batch", leave=False
                        ):
                            if not isinstance(data_list, ApiDataRetrievalFailed):
                                processed_items = [
                                    self.process_item(item)
                                    for item in data_list["products"]
                                ]
                                self._db_bulk_writer(processed_items)
                                pbar.update(len(data_list["products"]))
                            else:
                                self.logger.error(
                                    f"Retrieval of api data on url: {data_list.args[0]} failed...."
                                )

            # Set the last update time in the info collection
            self.setColUpdate(self.feed_type.lower(), self.last_modified)

        self.logger.info(
            "Duration: {}".format(datetime.timedelta(seconds=time.time() - start_time))
        )

    def update(self, **kwargs):
        self.logger.info("CPE database update started")

        # if collection is non-existent; assume it's not an update
        if self.feed_type.lower() not in self.getTableNames():
            DatabaseIndexer().create_indexes(collection=self.feed_type.lower())
            self.is_update = False

        self.process_downloads()

        self.logger.info("Finished CPE database update")

        return self.last_modified

    def populate(self, **kwargs):
        self.logger.info("CPE Database population started")

        self.is_update = False

        self.queue.clear()

        self.delColInfo(self.feed_type.lower())

        self.dropCollection(self.feed_type.lower())

        DatabaseIndexer().create_indexes(collection=self.feed_type.lower())

        self.process_downloads()

        self.logger.info("Finished CPE database population")

        return self.last_modified


class CVEDownloads(NVDApiHandler):
    """
    Class processing CVE source files
    """

    def __init__(self):
        self.feed_type = "CVES"

        super().__init__(self.feed_type)

        self.logger = logging.getLogger("CVEDownloads")

    @staticmethod
    def get_cve_year_range():
        """
        Method to fetch the years where we need cve's for
        """
        for a_year in range(cveStartYear, year):
            yield a_year

    @staticmethod
    def get_cpe_info(cpeuri):
        query = {}
        version_info = ""
        if "versionStartExcluding" in cpeuri:
            query["versionStartExcluding"] = cpeuri["versionStartExcluding"]
            version_info += query["versionStartExcluding"] + "_VSE"
        if "versionStartIncluding" in cpeuri:
            query["versionStartIncluding"] = cpeuri["versionStartIncluding"]
            version_info += query["versionStartIncluding"] + "_VSI"
        if "versionEndExcluding" in cpeuri:
            query["versionEndExcluding"] = cpeuri["versionEndExcluding"]
            version_info += query["versionEndExcluding"] + "_VEE"
        if "versionEndIncluding" in cpeuri:
            query["versionEndIncluding"] = cpeuri["versionEndIncluding"]
            version_info += query["versionEndIncluding"] + "_VEI"

        return query, version_info

    @staticmethod
    def add_if_missing(cve, key, value):
        if value not in cve[key]:
            cve[key].append(value)
        return cve

    @staticmethod
    def get_vendor_product(cpeUri):
        vendor = cpeUri.split(":")[3]
        product = cpeUri.split(":")[4]
        return vendor, product

    @staticmethod
    def stem(cpeUri):
        cpeArr = cpeUri.split(":")
        return ":".join(cpeArr[:5])

    def file_to_queue(self, *args):
        pass

    def process_the_item(self, item=None):
        if item is None:
            return None

        cve = {
            "id": item["cve"]["id"],
            "assigner": item["cve"]["sourceIdentifier"],
            "status": item["cve"]["vulnStatus"],
            "published": parse_datetime(item["cve"]["published"], ignoretz=True),
            "modified": parse_datetime(item["cve"]["lastModified"], ignoretz=True),
            "lastModified": parse_datetime(item["cve"]["lastModified"], ignoretz=True),
        }

        for description in item["cve"]["descriptions"]:
            if description["lang"] == "en":
                if "summary" in cve:
                    cve["summary"] += " {}".format(description["value"])
                else:
                    cve["summary"] = description["value"]

        if "metrics" in item["cve"]:
            cve["access"] = {}
            cve["impact"] = {}
            if "cvssMetricV3" in item["cve"]["metrics"]:
                cve["impact3"] = {}
                cve["exploitability3"] = {}
                cve["impact3"]["availability"] = item["cve"]["metrics"]["cvssMetricV3"][
                    0
                ]["cvssData"]["availabilityImpact"]
                cve["impact3"]["confidentiality"] = item["cve"]["metrics"][
                    "cvssMetricV3"
                ][0]["cvssData"]["confidentialityImpact"]
                cve["impact3"]["integrity"] = item["cve"]["metrics"]["cvssMetricV3"][0][
                    "cvssData"
                ]["integrityImpact"]
                cve["exploitability3"]["attackvector"] = item["cve"]["metrics"][
                    "cvssMetricV3"
                ][0]["cvssData"]["attackVector"]
                cve["exploitability3"]["attackcomplexity"] = item["cve"]["metrics"][
                    "cvssMetricV3"
                ][0]["cvssData"]["attackComplexity"]
                cve["exploitability3"]["privilegesrequired"] = item["cve"]["metrics"][
                    "cvssMetricV3"
                ][0]["cvssData"]["privilegesRequired"]
                cve["exploitability3"]["userinteraction"] = item["cve"]["metrics"][
                    "cvssMetricV3"
                ][0]["cvssData"]["userInteraction"]
                cve["exploitability3"]["scope"] = item["cve"]["metrics"][
                    "cvssMetricV3"
                ][0]["cvssData"]["scope"]
                cve["cvss3"] = float(
                    item["cve"]["metrics"]["cvssMetricV3"][0]["cvssData"]["baseScore"]
                )
                cve["cvss3Vector"] = item["cve"]["metrics"]["cvssMetricV3"][0][
                    "cvssData"
                ]["vectorString"]
                cve["impactScore3"] = float(
                    item["cve"]["metrics"]["cvssMetricV3"][0]["impactScore"]
                )
                cve["exploitabilityScore3"] = float(
                    item["cve"]["metrics"]["cvssMetricV3"][0]["exploitabilityScore"]
                )
                cve["cvss3Time"] = parse_datetime(
                    item["cve"]["lastModified"], ignoretz=True
                )
                cve["cvss3Type"] = item["cve"]["metrics"]["cvssMetricV3"][0]["type"]
                cve["cvss3Source"] = item["cve"]["metrics"]["cvssMetricV3"][0]["source"]
            else:
                cve["cvss3"] = None

            if "cvssMetricV2" in item["cve"]["metrics"]:
                cve["access"]["authentication"] = item["cve"]["metrics"][
                    "cvssMetricV2"
                ][0]["cvssData"]["authentication"]
                cve["access"]["complexity"] = item["cve"]["metrics"]["cvssMetricV2"][0][
                    "cvssData"
                ]["accessComplexity"]
                cve["access"]["vector"] = item["cve"]["metrics"]["cvssMetricV2"][0][
                    "cvssData"
                ]["accessVector"]
                cve["impact"]["availability"] = item["cve"]["metrics"]["cvssMetricV2"][
                    0
                ]["cvssData"]["availabilityImpact"]
                cve["impact"]["confidentiality"] = item["cve"]["metrics"][
                    "cvssMetricV2"
                ][0]["cvssData"]["confidentialityImpact"]
                cve["impact"]["integrity"] = item["cve"]["metrics"]["cvssMetricV2"][0][
                    "cvssData"
                ]["integrityImpact"]
                cve["cvss"] = float(
                    item["cve"]["metrics"]["cvssMetricV2"][0]["cvssData"]["baseScore"]
                )
                cve["exploitabilityScore"] = float(
                    item["cve"]["metrics"]["cvssMetricV2"][0]["exploitabilityScore"]
                )
                cve["impactScore"] = float(
                    item["cve"]["metrics"]["cvssMetricV2"][0]["impactScore"]
                )
                cve["cvssTime"] = parse_datetime(
                    item["cve"]["lastModified"], ignoretz=True
                )  # NVD JSON lacks the CVSS time which was present in the original XML format
                cve["cvssVector"] = item["cve"]["metrics"]["cvssMetricV2"][0][
                    "cvssData"
                ]["vectorString"]
                cve["cvssType"] = item["cve"]["metrics"]["cvssMetricV2"][0]["type"]
                cve["cvssSource"] = item["cve"]["metrics"]["cvssMetricV2"][0]["source"]
            else:
                cve["cvss"] = None

        if "references" in item["cve"]:
            cve["references"] = []
            for ref in item["cve"]["references"]:
                cve["references"].append(ref["url"])

        if "configurations" in item["cve"]:
            cve["vulnerable_configuration"] = []
            cve["vulnerable_product"] = []
            cve["vendors"] = []
            cve["products"] = []
            cve["vulnerable_product_stems"] = []
            cve["vulnerable_configuration_stems"] = []
            for cpe in item["cve"]["configurations"][0]["nodes"]:
                if "cpeMatch" in cpe:
                    for cpeuri in cpe["cpeMatch"]:
                        if "criteria" not in cpeuri:
                            continue
                        if cpeuri["vulnerable"]:
                            query, version_info = self.get_cpe_info(cpeuri)
                            if query != {}:
                                query["id"] = hashlib.sha1(
                                    cpeuri["criteria"].encode("utf-8")
                                    + version_info.encode("utf-8")
                                ).hexdigest()
                                cpe_info = self.getCPEVersionInformation(query)
                                if cpe_info:
                                    if cpe_info["cpeMatch"]:
                                        for vulnerable_version in cpe_info["cpeMatch"]:
                                            cve = self.add_if_missing(
                                                cve,
                                                "vulnerable_product",
                                                vulnerable_version["criteria"],
                                            )
                                            cve = self.add_if_missing(
                                                cve,
                                                "vulnerable_configuration",
                                                vulnerable_version["criteria"],
                                            )
                                            cve = self.add_if_missing(
                                                cve,
                                                "vulnerable_configuration_stems",
                                                self.stem(
                                                    vulnerable_version["criteria"]
                                                ),
                                            )
                                            vendor, product = self.get_vendor_product(
                                                vulnerable_version["criteria"]
                                            )
                                            cve = self.add_if_missing(
                                                cve, "vendors", vendor
                                            )
                                            cve = self.add_if_missing(
                                                cve, "products", product
                                            )
                                            cve = self.add_if_missing(
                                                cve,
                                                "vulnerable_product_stems",
                                                self.stem(
                                                    vulnerable_version["criteria"]
                                                ),
                                            )
                                    else:
                                        cve = self.add_if_missing(
                                            cve,
                                            "vulnerable_product",
                                            cpeuri["criteria"],
                                        )
                                        cve = self.add_if_missing(
                                            cve,
                                            "vulnerable_configuration",
                                            cpeuri["criteria"],
                                        )
                                        cve = self.add_if_missing(
                                            cve,
                                            "vulnerable_configuration_stems",
                                            self.stem(cpeuri["criteria"]),
                                        )
                                        vendor, product = self.get_vendor_product(
                                            cpeuri["criteria"]
                                        )
                                        cve = self.add_if_missing(
                                            cve, "vendors", vendor
                                        )
                                        cve = self.add_if_missing(
                                            cve, "products", product
                                        )
                                        cve = self.add_if_missing(
                                            cve,
                                            "vulnerable_product_stems",
                                            self.stem(cpeuri["criteria"]),
                                        )
                            else:
                                # If the cpeMatch did not have any of the version start/end modifiers,
                                # add the CPE string as it is.
                                cve = self.add_if_missing(
                                    cve, "vulnerable_product", cpeuri["criteria"]
                                )
                                cve = self.add_if_missing(
                                    cve, "vulnerable_configuration", cpeuri["criteria"]
                                )
                                cve = self.add_if_missing(
                                    cve,
                                    "vulnerable_configuration_stems",
                                    self.stem(cpeuri["criteria"]),
                                )
                                vendor, product = self.get_vendor_product(
                                    cpeuri["criteria"]
                                )
                                cve = self.add_if_missing(cve, "vendors", vendor)
                                cve = self.add_if_missing(cve, "products", product)
                                cve = self.add_if_missing(
                                    cve,
                                    "vulnerable_product_stems",
                                    self.stem(cpeuri["criteria"]),
                                )
                        else:
                            cve = self.add_if_missing(
                                cve, "vulnerable_configuration", cpeuri["criteria"]
                            )
                            cve = self.add_if_missing(
                                cve,
                                "vulnerable_configuration_stems",
                                self.stem(cpeuri["criteria"]),
                            )
                if "children" in cpe:
                    for child in cpe["children"]:
                        if "cpeMatch" in child:
                            for cpeuri in child["cpeMatch"]:
                                if "criteria" not in cpeuri:
                                    continue
                                if cpeuri["vulnerable"]:
                                    query, version_info = self.get_cpe_info(cpeuri)
                                    if query != {}:
                                        query["id"] = hashlib.sha1(
                                            cpeuri["criteria"].encode("utf-8")
                                            + version_info.encode("utf-8")
                                        ).hexdigest()
                                        cpe_info = self.getCPEVersionInformation(query)
                                        if cpe_info:
                                            if cpe_info["cpeMatch"]:
                                                for vulnerable_version in cpe_info[
                                                    "cpeMatch"
                                                ]:
                                                    cve = self.add_if_missing(
                                                        cve,
                                                        "vulnerable_product",
                                                        vulnerable_version["criteria"],
                                                    )
                                                    cve = self.add_if_missing(
                                                        cve,
                                                        "vulnerable_configuration",
                                                        vulnerable_version["criteria"],
                                                    )
                                                    cve = self.add_if_missing(
                                                        cve,
                                                        "vulnerable_configuration_stems",
                                                        self.stem(
                                                            vulnerable_version[
                                                                "criteria"
                                                            ]
                                                        ),
                                                    )
                                                    (
                                                        vendor,
                                                        product,
                                                    ) = self.get_vendor_product(
                                                        vulnerable_version["criteria"]
                                                    )
                                                    cve = self.add_if_missing(
                                                        cve, "vendors", vendor
                                                    )
                                                    cve = self.add_if_missing(
                                                        cve, "products", product
                                                    )
                                                    cve = self.add_if_missing(
                                                        cve,
                                                        "vulnerable_product_stems",
                                                        self.stem(
                                                            vulnerable_version[
                                                                "criteria"
                                                            ]
                                                        ),
                                                    )
                                            else:
                                                cve = self.add_if_missing(
                                                    cve,
                                                    "vulnerable_product",
                                                    cpeuri["criteria"],
                                                )
                                                cve = self.add_if_missing(
                                                    cve,
                                                    "vulnerable_configuration",
                                                    cpeuri["criteria"],
                                                )
                                                cve = self.add_if_missing(
                                                    cve,
                                                    "vulnerable_configuration_stems",
                                                    self.stem(cpeuri["criteria"]),
                                                )
                                                (
                                                    vendor,
                                                    product,
                                                ) = self.get_vendor_product(
                                                    cpeuri["criteria"]
                                                )
                                                cve = self.add_if_missing(
                                                    cve, "vendors", vendor
                                                )
                                                cve = self.add_if_missing(
                                                    cve, "products", product
                                                )
                                                cve = self.add_if_missing(
                                                    cve,
                                                    "vulnerable_product_stems",
                                                    self.stem(cpeuri["criteria"]),
                                                )
                                    else:
                                        # If the cpeMatch did not have any of the version start/end modifiers,
                                        # add the CPE string as it is.
                                        if "criteria" not in cpeuri:
                                            continue
                                        cve = self.add_if_missing(
                                            cve,
                                            "vulnerable_product",
                                            cpeuri["criteria"],
                                        )
                                        cve = self.add_if_missing(
                                            cve,
                                            "vulnerable_configuration",
                                            cpeuri["criteria"],
                                        )
                                        cve = self.add_if_missing(
                                            cve,
                                            "vulnerable_configuration_stems",
                                            self.stem(cpeuri["criteria"]),
                                        )
                                        vendor, product = self.get_vendor_product(
                                            cpeuri["criteria"]
                                        )
                                        cve = self.add_if_missing(
                                            cve, "vendors", vendor
                                        )
                                        cve = self.add_if_missing(
                                            cve, "products", product
                                        )
                                        cve = self.add_if_missing(
                                            cve,
                                            "vulnerable_product_stems",
                                            self.stem(cpeuri["criteria"]),
                                        )
                                else:
                                    if "criteria" not in cpeuri:
                                        continue
                                    cve = self.add_if_missing(
                                        cve,
                                        "vulnerable_configuration",
                                        cpeuri["criteria"],
                                    )
                                    cve = self.add_if_missing(
                                        cve,
                                        "vulnerable_configuration_stems",
                                        self.stem(cpeuri["criteria"]),
                                    )

        if "weaknesses" in item["cve"]:
            for problem in item["cve"]["weaknesses"]:
                for cwe in problem[
                    "description"
                ]:  # NVD JSON not clear if we can get more than one CWE per CVE (until we take the last one) -
                    # NVD-CWE-Other??? list?
                    if cwe["lang"] == "en":
                        cve["cwe"] = cwe["value"]
            if not ("cwe" in cve):
                cve["cwe"] = defaultvalue["cwe"]
        else:
            cve["cwe"] = defaultvalue["cwe"]

        cve["vulnerable_configuration_cpe_2_2"] = []

        return cve

    def process_downloads(self, sites=None):
        """
        Method to download and process files
        """

        self.logger.info("Starting download...")

        start_time = time.time()

        self.last_modified = datetime.datetime.now()

        self.logger.debug(
            f"do_process = {self.do_process}; is_update = {self.is_update}"
        )

        if self.do_process:

            if not self.is_update:

                total_results = self.api_handler.get_count(
                    self.api_handler.datasource.CVE
                )

                self.logger.info(f"Preparing to download {total_results} CVE entries")

                with tqdm(
                    desc="Downloading and processing content",
                    total=total_results,
                    position=0,
                    leave=True,
                ) as pbar:
                    for entry in self.api_handler.get_all_data(data_type="cve"):
                        # do something here with the results...
                        for data_list in tqdm(
                            entry, desc=f"Processing batch", leave=False
                        ):
                            if not isinstance(data_list, ApiDataRetrievalFailed):
                                processed_items = [
                                    self.process_item(item)
                                    for item in data_list["vulnerabilities"]
                                ]
                                self._db_bulk_writer(processed_items)
                                pbar.update(len(data_list["vulnerabilities"]))
                            else:
                                self.logger.error(
                                    f"Retrieval of api data on url: {data_list.args[0]} failed...."
                                )
            else:
                last_mod_start_date = self.database[self.feed_type.lower()].find_one(
                    {}, {"lastModified": 1}, sort=[("lastModified", -1)]
                )

                if "lastModified" in last_mod_start_date:
                    last_mod_start_date = last_mod_start_date["lastModified"]
                else:
                    raise KeyError(
                        "Missing field 'lastModified' from database query..."
                    )

                # Get datetime from runtime
                last_mod_end_date = datetime.datetime.now()

                total_results = self.api_handler.get_count(
                    self.api_handler.datasource.CVE,
                    last_mod_start_date=last_mod_start_date,
                    last_mod_end_date=last_mod_end_date,
                )

                self.logger.info(f"Preparing to download {total_results} CVE entries")

                with tqdm(
                    desc="Downloading and processing content",
                    total=total_results,
                    position=0,
                    leave=True,
                ) as pbar:
                    for entry in self.api_handler.get_all_data(
                        data_type="cve",
                        last_mod_start_date=last_mod_start_date,
                        last_mod_end_date=last_mod_end_date,
                    ):
                        # do something here with the results...
                        for data_list in tqdm(
                            entry, desc=f"Processing batch", leave=False
                        ):
                            if not isinstance(data_list, ApiDataRetrievalFailed):
                                processed_items = [
                                    self.process_item(item)
                                    for item in data_list["vulnerabilities"]
                                ]
                                self._db_bulk_writer(processed_items)
                                pbar.update(len(data_list["vulnerabilities"]))
                            else:
                                self.logger.error(
                                    f"Retrieval of api data on url: {data_list.args[0]} failed...."
                                )

            # Set the last update time in the info collection
            self.setColUpdate(self.feed_type.lower(), self.last_modified)

        self.logger.info(
            "Duration: {}".format(datetime.timedelta(seconds=time.time() - start_time))
        )

    def update(self):
        self.logger.info("CVE database update started")

        # if collection is non-existent; assume it's not an update
        if self.feed_type.lower() not in self.getTableNames():
            DatabaseIndexer().create_indexes(collection=self.feed_type.lower())
            self.is_update = False

        self.process_downloads()

        self.logger.info("Finished CVE database update")

        return self.last_modified

    def populate(self):

        self.logger.info("CVE database population started")

        self.logger.info(
            "Starting CVE database population starting from year: {}".format(
                cveStartYear
            )
        )

        self.is_update = False

        self.queue.clear()

        self.delColInfo(self.feed_type.lower())

        self.dropCollection(self.feed_type.lower())

        DatabaseIndexer().create_indexes(collection=self.feed_type.lower())

        self.process_downloads()

        self.logger.info("Finished CVE database population")

        return self.last_modified


class VIADownloads(JSONFileHandler):
    """
    Class processing VIA4 source files
    """

    def __init__(self):
        self.feed_type = "VIA4"
        self.prefix = "cves"
        super().__init__(self.feed_type, self.prefix)

        self.feed_url = Configuration.getFeedURL(self.feed_type.lower())

        self.logger = logging.getLogger("VIADownloads")

    def file_to_queue(self, file_tuple):

        working_dir, filename = file_tuple

        for cve in self.ijson_handler.fetch(filename=filename, prefix=self.prefix):
            x = 0
            for key, val in cve.items():
                entry_dict = {"id": key}
                entry_dict.update(val)
                self.process_item(item=entry_dict)
                x += 1

            self.logger.debug("Processed {} items from file: {}".format(x, filename))

        with open(filename, "rb") as input_file:
            data = json.loads(input_file.read().decode("utf-8"))

            self.setColInfo("via4", "sources", data["metadata"]["sources"])
            self.setColInfo("via4", "searchables", data["metadata"]["searchables"])

            self.logger.debug("Processed metadata from file: {}".format(filename))

        try:
            self.logger.debug("Removing working dir: {}".format(working_dir))
            shutil.rmtree(working_dir)
        except Exception as err:
            self.logger.error(
                "Failed to remove working dir; error produced: {}".format(err)
            )

    def process_item(self, item):

        if self.is_update:
            self.queue.put(
                DatabaseAction(
                    action=DatabaseAction.actions.UpdateOne,
                    collection=self.feed_type.lower(),
                    doc=item,
                )
            )
        else:
            self.queue.put(
                DatabaseAction(
                    action=DatabaseAction.actions.InsertOne,
                    collection=self.feed_type.lower(),
                    doc=item,
                )
            )

    def update(self, **kwargs):
        self.logger.info("VIA4 database update started")

        # if collection is non-existent; assume it's not an update
        if self.feed_type.lower() not in self.getTableNames():
            self.is_update = False

        self.process_downloads([self.feed_url])

        self.logger.info("Finished VIA4 database update")

        return self.last_modified

    def populate(self, **kwargs):
        self.is_update = False
        self.queue.clear()

        self.delColInfo(self.feed_type.lower())

        self.dropCollection(self.feed_type.lower())

        return self.update()


class CAPECDownloads(XMLFileHandler):
    """
    Class processing CAPEC source files
    """

    def __init__(self):
        self.feed_type = "CAPEC"
        super().__init__(self.feed_type)

        self.feed_url = Configuration.getFeedURL(self.feed_type.lower())

        self.logger = logging.getLogger("CAPECDownloads")

        # make parser
        self.parser = make_parser()
        self.ch = CapecHandler()
        self.parser.setContentHandler(self.ch)

    def file_to_queue(self, file_tuple):

        working_dir, filename = file_tuple

        self.parser.parse(filename)
        x = 0
        for attack in self.ch.capec:
            self.process_item(attack)
            x += 1

        self.logger.debug("Processed {} entries from file: {}".format(x, filename))

        try:
            self.logger.debug("Removing working dir: {}".format(working_dir))
            shutil.rmtree(working_dir)
        except Exception as err:
            self.logger.error(
                "Failed to remove working dir; error produced: {}".format(err)
            )

    def update(self, **kwargs):
        self.logger.info("CAPEC database update started")

        # if collection is non-existent; assume it's not an update
        if self.feed_type.lower() not in self.getTableNames():
            self.is_update = False

        self.process_downloads([self.feed_url])

        self.logger.info("Finished CAPEC database update")

        return self.last_modified

    def populate(self, **kwargs):
        self.is_update = False
        self.queue.clear()

        self.delColInfo(self.feed_type.lower())

        self.dropCollection(self.feed_type.lower())

        return self.update()


class CWEDownloads(XMLFileHandler):
    """
    Class processing CWE source files
    """

    def __init__(self):
        self.feed_type = "CWE"
        super().__init__(self.feed_type)

        self.feed_url = Configuration.getFeedURL(self.feed_type.lower())

        self.logger = logging.getLogger("CWEDownloads")

        # make parser
        self.parser = make_parser()
        self.ch = CWEHandler()
        self.parser.setContentHandler(self.ch)

    def file_to_queue(self, file_tuple):

        working_dir, filename = file_tuple

        for f in glob.glob(f"{working_dir}/*.xml"):
            filename = f

        self.parser.parse(f"file://{filename}")
        x = 0
        for cwe in self.ch.cwe:
            try:
                cwe["related_weaknesses"] = list(set(cwe["related_weaknesses"]))
            except KeyError:
                pass
            self.process_item(cwe)
            x += 1

        self.logger.debug("Processed {} entries from file: {}".format(x, filename))

        try:
            self.logger.debug("Removing working dir: {}".format(working_dir))
            shutil.rmtree(working_dir)
        except Exception as err:
            self.logger.error(
                "Failed to remove working dir; error produced: {}".format(err)
            )

    def update(self, **kwargs):
        self.logger.info("CWE database update started")

        # if collection is non-existent; assume it's not an update
        if self.feed_type.lower() not in self.getTableNames():
            self.is_update = False

        self.process_downloads([self.feed_url])

        self.logger.info("Finished CWE database update")

        return self.last_modified

    def populate(self, **kwargs):
        self.is_update = False
        self.queue.clear()

        self.delColInfo(self.feed_type.lower())

        self.dropCollection(self.feed_type.lower())

        return self.update()


MongoUniqueIndex = namedtuple("MongoUniqueIndex", "index name unique")
MongoAddIndex = namedtuple("MongoAddIndex", "index name")


class DatabaseIndexer(object):
    """
    Class processing the Mongodb indexes
    """

    def __init__(self):

        database = MongoDBConnection(**json.loads(os.getenv("MONGODB_CON_DETAILS")))
        self.database = database._dbclient

        self.indexes = {
            "cpe": [
                MongoUniqueIndex(index=[("id", ASCENDING)], name="id", unique=True),
                MongoAddIndex(index=[("vendor", ASCENDING)], name="vendor"),
                MongoAddIndex(index=[("product", ASCENDING)], name="product"),
            ],
            "cpeother": [
                MongoUniqueIndex(index=[("id", ASCENDING)], name="id", unique=True)
            ],
            "cves": [
                MongoAddIndex(index=[("id", ASCENDING)], name="id"),
                MongoAddIndex(
                    index=[("vulnerable_configuration", ASCENDING)],
                    name="vulnerable_configuration",
                ),
                MongoAddIndex(
                    index=[("vulnerable_product", ASCENDING)], name="vulnerable_product"
                ),
                MongoAddIndex(index=[("modified", ASCENDING)], name="modified"),
                MongoAddIndex(index=[("published", ASCENDING)], name="published"),
                MongoAddIndex(index=[("lastModified", ASCENDING)], name="lastModified"),
                MongoAddIndex(index=[("cvss", ASCENDING)], name="cvss"),
                MongoAddIndex(index=[("cvss3", ASCENDING)], name="cvss3"),
                MongoAddIndex(index=[("summary", TEXT)], name="summary"),
                MongoAddIndex(index=[("vendors", ASCENDING)], name="vendors"),
                MongoAddIndex(index=[("products", ASCENDING)], name="products"),
                MongoAddIndex(
                    index=[("vulnerable_product_stems", ASCENDING)],
                    name="vulnerable_product_stems",
                ),
                MongoAddIndex(
                    index=[("vulnerable_configuration_stems", ASCENDING)],
                    name="vulnerable_configuration_stems",
                ),
            ],
            "via4": [MongoAddIndex(index=[("id", ASCENDING)], name="id")],
            "mgmt_whitelist": [MongoAddIndex(index=[("id", ASCENDING)], name="id")],
            "mgmt_blacklist": [MongoAddIndex(index=[("id", ASCENDING)], name="id")],
            "capec": [
                MongoAddIndex(
                    index=[("related_weakness", ASCENDING)], name="related_weakness"
                )
            ],
        }

        self.logger = logging.getLogger("DatabaseIndexer")

    def getInfo(self, collection):
        return self.sanitize(self.database["info"].find_one({"db": collection}))

    def sanitize(self, x):
        if type(x) == pymongo.cursor.Cursor:
            x = list(x)
        if type(x) == list:
            for y in x:
                self.sanitize(y)
        if x and "_id" in x:
            x.pop("_id")
        return x

    def create_indexes(self, collection=None):

        if collection is not None:
            try:
                for each in self.indexes[collection]:
                    if isinstance(each, MongoUniqueIndex):
                        self.setIndex(
                            collection, each.index, name=each.name, unique=each.unique
                        )
                    elif isinstance(each, MongoAddIndex):
                        self.setIndex(collection, each.index, name=each.name)
            except KeyError:
                # no specific index given, continue
                self.logger.warning(
                    "Could not find the requested collection: {}, skipping...".format(
                        collection
                    )
                )
                pass

        else:
            for index in self.iter_indexes():
                self.setIndex(index[0], index[1])

            for collection in self.indexes.keys():
                for each in self.indexes[collection]:
                    if isinstance(each, MongoUniqueIndex):
                        self.setIndex(
                            collection, each.index, name=each.name, unique=each.unique
                        )
                    elif isinstance(each, MongoAddIndex):
                        self.setIndex(collection, each.index, name=each.name)

    def iter_indexes(self):
        for each in self.get_via4_indexes():
            yield each

    def get_via4_indexes(self):
        via4 = self.getInfo("via4")
        result = []
        if via4:
            for index in via4.get("searchables", []):
                result.append(("via4", index))
        return result

    def setIndex(self, col, field, **kwargs):
        try:
            self.database[col].create_index(field, **kwargs)
            self.logger.info("Success to create index %s on %s" % (field, col))
        except Exception as e:
            self.logger.error("Failed to create index %s on %s: %s" % (col, field, e))



"""
cpe
===
"""
import re

from pymongo import DESCENDING

from CveXplore.common.cpe_converters import from2to3CPE
from CveXplore.common.data_source_connection import DatasourceConnection


class Cpe(DatasourceConnection):
    """
    Cpe database object
    """

    def __init__(self, **kwargs):

        super().__init__("cpe")

        for each in kwargs:
            setattr(self, each, kwargs[each])

    def iter_cves_matching_cpe(self, vuln_prod_search=False):
        """
        Generator function for iterating over cve's matching this CPE. By default the search will be made matching
        the configuration fields of the cves documents.

        :param vuln_prod_search: Search for matching products instead of configurations
        :type vuln_prod_search: bool
        :return: Matching CVES
        :rtype: Cves
        """

        cpe_searchField = (
            "vulnerable_product" if vuln_prod_search else "vulnerable_configuration"
        )

        # format to cpe2.3
        cpe_string = from2to3CPE(self.cpe_2_2)

        if cpe_string.startswith("cpe"):
            # strict search with term starting with cpe; e.g: cpe:2.3:o:microsoft:windows_7:*:sp1:*:*:*:*:*:*

            remove_trailing_regex_stars = r"(?:\:|\:\:|\:\*)+$"

            cpe_regex = re.escape(re.sub(remove_trailing_regex_stars, "", cpe_string))

            cpe_regex_string = r"^{}:".format(cpe_regex)
        else:
            # more general search on same field; e.g. microsoft:windows_7
            cpe_regex_string = "{}".format(re.escape(cpe_string))

        results = self._datasource_connection.store_cves.find(
            {cpe_searchField: {"$regex": cpe_regex_string}}
        ).sort("cvss", DESCENDING)

        for each in results:
            if each is not None:
                yield each
            else:
                yield None

    def iter_cpe_names(self):
        """
        Generator function for iterating over cpe_names for this CPE.

        :return: cpe_name
        :rtype: str
        """

        if hasattr(self, "cpe_name"):
            for each in self.cpe_name:
                yield each
        else:
            return "None"

    def to_dict(self):
        """
        Method to convert the entire object to a dictionary

        :return: Data from object
        :rtype: dict
        """

        return {k: v for (k, v) in self.__dict__.items() if not k.startswith("_")}

    def __eq__(self, other):
        return self.__dict__ == other.__dict__

    def __ne__(self, other):
        return self.__dict__ != other.__dict__

    def __repr__(self):
        """String representation of object"""
        return "<< Cpe:{} >>".format(self.id)



"""
cves
====
"""
import datetime

from CveXplore.common.data_source_connection import DatasourceConnection


class Cves(DatasourceConnection):
    """
    Cves database object
    """

    def __init__(self, **kwargs):

        super().__init__("cves")

        for each in kwargs:
            setattr(self, each, kwargs[each])

        if hasattr(self, "cwe"):
            if isinstance(self.cwe, str):
                if self.cwe.lower() != "unknown":
                    cwe_id = self.cwe[4:]
                    # check if cwe_id can be cast to int
                    try:
                        if int(cwe_id):
                            results = getattr(
                                self._datasource_connection, "store_{}".format("cwe")
                            ).find_one({"id": cwe_id})
                            if results is not None:
                                self.cwe = results
                    except ValueError:
                        pass

                    capecs = self._datasource_connection.store_capec.find(
                        {"related_weakness": {"$in": [cwe_id]}}
                    )

                    setattr(self, "capec", list(capecs))

        via4s = self._datasource_connection.store_via4.find_one({"id": self.id})

        if via4s is not None:
            setattr(self, "via4_references", via4s)

    def iter_vuln_configurations(self):
        """
        Generator function for iterating over vulnerable configurations for this CVE.

        :return: vulnerable_configuration
        :rtype: str
        """

        if hasattr(self, "vulnerable_configuration"):
            for each in self.vulnerable_configuration:
                yield each

    def iter_references(self):
        """
        Generator function for iterating over references for this CVE.

        :return: references
        :rtype: str
        """

        if hasattr(self, "references"):
            for each in self.references:
                yield each

    def iter_capec(self):
        """
        Generator function for iterating over capecs for this CVE.

        :return: capec
        :rtype: Capec
        """

        if hasattr(self, "capec"):
            for each in self.capec:
                yield each

    def to_dict(self):
        """
        Method to convert the entire object to a dictionary

        :return: Data from object
        :rtype: dict
        """

        full_dict = {k: v for (k, v) in self.__dict__.items() if not k.startswith("_")}

        for k, v in full_dict.items():
            if isinstance(v, DatasourceConnection):
                full_dict[k] = v.to_dict()
            if isinstance(v, list):
                if len(v) > 0:
                    if isinstance(v[0], DatasourceConnection):
                        new_list = []
                        for each in v:
                            new_list.append(each.to_dict())
                        full_dict[k] = new_list
            if isinstance(v, datetime.datetime):
                full_dict[k] = str(v)

        return full_dict

    def __eq__(self, other):
        return self.__dict__ == other.__dict__

    def __ne__(self, other):
        return self.__dict__ != other.__dict__

    def __repr__(self):
        """String representation of object"""
        return "<< Cves:{} >>".format(self.id)



"""
Specific database functions
===========================
"""
from pymongo import DESCENDING

from CveXplore.database.helpers.generic_db import GenericDatabaseFactory


class CvesDatabaseFunctions(GenericDatabaseFactory):
    """
    The CvesDatabaseFunctions is a specific class that provides the cves attribute of a CveXplore instance additional
    functions that only apply to the 'cves' collection
    """

    def __init__(self, collection):
        super().__init__(collection)

    def get_cves_for_vendor(self, vendor, limit=0):
        """
        Function to return cves based on a given vendor. By default to result is sorted descending on th cvss field.

        :param vendor: A vendor to search for; e.g. microsoft
        :type vendor: str
        :param limit: Limit the amount of returned results
        :type limit: int
        :return: List with cves objects
        :rtype: list
        """

        return list(
            self._datasource_collection_connection.find({"vendors": vendor})
            .limit(limit)
            .sort("cvss", DESCENDING)
        )

    def __repr__(self):
        """String representation of object"""
        return "<< CvesDatabaseFunctions:{} >>".format(self._collection)



"""
Generic database functions
==========================
"""
import re

from CveXplore.common.data_source_connection import DatasourceConnection
from CveXplore.common.db_mapping import database_mapping


class GenericDatabaseFactory(DatasourceConnection):
    """
    The GenericDatabaseFactory handles the creation of general, collection based, functions which provide an instance
    of CveXplore functions that apply to the given collection.
    """

    def __init__(self, collection):
        """
        Create a new GenericDatabaseFactory and create field specific functions based on the __default_fields and
        the __fields_mapping.

        :param collection: Collection to create the functions for
        :type collection: str
        """
        super().__init__(collection=collection)

        self.__database_mapping = database_mapping

        self.__default_fields = ["id"]

        self.__fields_mapping = {
            "capec": [
                "name",
                "summary",
                "prerequisites",
                "solutions",
                "loa",
                "typical_severity",
            ],
            "cpe": ["title", "cpe_2_2", "vendor", "product"],
            "cwe": ["name", "status", "Description"],
            "via4": ["name"],
            "cves": [
                "cvss",
                "cvss3",
                "summary",
                "vendors",
                "products",
                "lastModified",
                "Modified",
                "Published",
            ],
        }

        total_fields_list = (
            self.__default_fields + self.__fields_mapping[self._collection]
        )
        for field in total_fields_list:
            setattr(
                self,
                field,
                GenericDatabaseFieldsFunctions(
                    field=field, collection=self._collection
                ),
            )

    def get_by_id(self, id):
        """
        Method to fetch a specific collection entry via it's id number

        :param id: Id number
        :type id: str
        :return: Requested data or None
        :rtype: object
        """

        if not isinstance(id, str):
            try:
                id = str(id)
            except ValueError:
                return "Provided value is not a string nor can it be cast to one"

        return self._datasource_collection_connection.find_one({"id": id})

    def __repr__(self):
        """String representation of object"""
        return "<< GenericDatabaseFactory:{} >>".format(self._collection)


class GenericDatabaseFieldsFunctions(DatasourceConnection):
    """
    The GenericDatabaseFieldsFunctions handles the creation of general, field based, functions
    """

    def __init__(self, field, collection):
        """
        Create a new GenericDatabaseFieldsFunctions and create field specific functions.

        :param field: Field name
        :type field: str
        :param collection: Collection to create the field functions for
        :type collection: str
        """
        super().__init__(collection=collection)

        self.__field = field

    def search(self, value):
        """
        Method for searching for a given value. The value shall be converted to a regex.

        :param value: Search for value
        :type value: str
        :return: Data or None
        :rtype: object
        """

        regex = re.compile(value, re.IGNORECASE)

        query = {self.__field: {"$regex": regex}}

        return self._datasource_collection_connection.find(query)

    def find(self, value=None):
        """
        Method to find a given value.

        :param value: Find a value
        :type value: str
        :return: Data or None
        :rtype: object
        """

        if value is not None:
            query = {self.__field: value}
        else:
            query = None

        return self._datasource_collection_connection.find(query)

    def __repr__(self):
        """String representation of object"""
        return "<< GenericDatabaseFieldsFunctions:{} >>".format(self._collection)



"""
Configuration
=============
"""
import ast
import datetime
import json
import os
import re
import shutil
from json import JSONDecodeError

from dotenv import load_dotenv

if not os.path.exists(os.path.expanduser("~/.cvexplore")):
    os.mkdir(os.path.expanduser("~/.cvexplore"))

user_wd = os.path.expanduser("~/.cvexplore")

if not os.path.exists(os.path.join(user_wd, ".env")):
    shutil.copyfile(
        os.path.join(os.path.dirname(__file__), ".env_example"),
        os.path.join(user_wd, ".env"),
    )

load_dotenv(os.path.join(user_wd, ".env"))

if not os.path.exists(os.path.join(user_wd, ".sources.ini")):
    shutil.copyfile(
        os.path.join(os.path.dirname(__file__), ".sources.ini"),
        os.path.join(user_wd, ".sources.ini"),
    )


def getenv_bool(name: str, default: str = "False"):
    raw = os.getenv(name, default).title()
    return ast.literal_eval(raw)


def getenv_dict(name: str, default: dict = None):
    if default is None:
        default = {}

    raw = os.getenv(name, default)

    if not isinstance(raw, dict):
        try:
            the_dict = json.loads(raw)
            return the_dict
        except JSONDecodeError:
            raise

    return default


class Configuration(object):
    """
    Class holding the configuration
    """

    USER_HOME_DIR = user_wd

    CVE_START_YEAR = os.getenv("CVE_START_YEAR", 2002)

    if os.getenv("SOURCES") is not None:
        SOURCES = getenv_dict("SOURCES", None)
    else:
        with open(os.path.join(user_wd, ".sources.ini")) as f:
            SOURCES = json.loads(f.read())

    NVD_NIST_API_KEY = os.getenv("NVD_NIST_API_KEY", None)
    NVD_NIST_NO_REJECTED = getenv_bool("NVD_NIST_NO_REJECTED", "True")

    DEFAULT_SOURCES = {
        "cwe": "https://cwe.mitre.org/data/xml/cwec_latest.xml.zip",
        "capec": "https://capec.mitre.org/data/xml/capec_latest.xml",
        "via4": "https://www.cve-search.org/feeds/via4.json",
    }

    HTTP_PROXY = os.getenv("HTTP_PROXY", "")

    LOGGING_TO_FILE = getenv_bool("LOGGING_TO_FILE", "True")
    LOGGING_FILE_PATH = os.getenv("LOGGING_FILE_PATH", os.path.join(user_wd, "log"))

    if not os.path.exists(LOGGING_FILE_PATH):
        os.mkdir(LOGGING_FILE_PATH)

    LOGGING_MAX_FILE_SIZE = os.getenv("LOGGING_MAX_FILE_SIZE", "100MB")
    LOGGING_BACKLOG = os.getenv("LOGGING_BACKLOG", 5)
    LOGGING_FILE_NAME = os.getenv("LOGGING_FILE_NAME", "./cvexplore.log")
    LOGGING_UPDATE_FILE_NAME = os.getenv("LOGGING_FILE_NAME", "./update_populate.log")

    @classmethod
    def getCVEStartYear(cls):
        next_year = datetime.datetime.now().year + 1
        start_year = cls.CVE_START_YEAR
        if start_year < cls.CVE_START_YEAR or start_year > next_year:
            print(
                "The year %i is not a valid year.\ndefault year %i will be used."
                % (start_year, cls.default["CVEStartYear"])
            )
            start_year = cls.default["CVEStartYear"]
        return start_year

    @classmethod
    def getProxy(cls):
        return cls.HTTP_PROXY

    @classmethod
    def getFeedURL(cls, source):
        return cls.SOURCES[source]

    @classmethod
    def getUpdateLogFile(cls):
        return os.path.join(cls.LOGGING_FILE_PATH, cls.LOGGING_UPDATE_FILE_NAME)

    @classmethod
    def getMaxLogSize(cls):
        size = cls.LOGGING_MAX_FILE_SIZE
        split = re.findall("\d+|\D+", size)
        multipliers = {"KB": 1024, "MB": 1024 * 1024, "GB": 1024 * 1024 * 1024}
        if len(split) == 2:
            base = int(split[0])
            unit = split[1].strip().upper()
            return base * multipliers.get(unit, 1024 * 1024)
        # if size is not a correctly defined set it to 100MB
        else:
            return 100 * 1024 * 1024

    @classmethod
    def getBacklog(cls):
        return cls.LOGGING_BACKLOG



"""
CPE Converters
==============
"""
from CveXplore.database.helpers.cpe_conversion import cpe_uri_to_fs, cpe_fs_to_uri


def from2to3CPE(cpe, autofill=False):
    """
    Method to transform cpe2.2 to cpe2.3 format

    :param cpe: cpe2.2 string
    :type cpe: str
    :param autofill: Whether to cpe string should be autofilled with double quotes and hyphens
    :type autofill: bool
    :return: cpe2.3 formatted string
    :rtype: str
    """
    cpe = cpe.strip()
    if not cpe.startswith("cpe:2.3:"):
        if not cpe.startswith("cpe:/"):
            return False
        cpe = cpe_uri_to_fs(cpe)
    if autofill:
        e = cpe.split(":")
        for x in range(0, 13 - len(e)):
            cpe += ":-"
    return cpe


def from3to2CPE(cpe):
    """
    Method to transform cpe2.3 to cpe2.2 format

    :param cpe: cpe2.3 string
    :type cpe: str
    :return: cpe2.2 string
    :rtype: str
    """
    cpe = cpe.strip()
    if not cpe.startswith("cpe:/"):
        if not cpe.startswith("cpe:2.3:"):
            return False
        cpe = cpe_fs_to_uri(cpe)
    return cpe


def pad(seq, target_length, padding=None):
    length = len(seq)
    if length > target_length:
        return seq
    seq.extend([padding] * (target_length - length))
    return seq



import re

import dateutil.parser
from dateutil import tz

from CveXplore.database.helpers import cpe_conversion


def currentTime(utc):
    timezone = tz.tzlocal()
    utc = dateutil.parser.parse(utc)
    output = utc.astimezone(timezone)
    output = output.strftime("%d-%m-%Y - %H:%M")
    return output


def isURL(string):
    urlTypes = [re.escape(x) for x in ["http://", "https://", "www."]]
    return re.match("^(" + "|".join(urlTypes) + ")", string)


def vFeedName(string):
    string = string.replace("map_", "")
    string = string.replace("cve_", "")
    return string.title()


def mergeSearchResults(database, plugins):
    if "errors" in database:
        results = {"data": [], "errors": database["errors"]}
    else:
        results = {"data": []}

    data = []
    data.extend(database["data"])
    data.extend(plugins["data"])
    for cve in data:
        if not any(cve["id"] == entry["id"] for entry in results["data"]):
            results["data"].append(cve)
    return results


def tk_compile(regexes):
    if type(regexes) not in [list, tuple]:
        regexes = [regexes]
    r = []
    for rule in regexes:
        r.append(re.compile(rule))
    return r


# Convert cpe2.2 url encoded to cpe2.3 char escaped
# cpe:2.3:o:cisco:ios:12.2%281%29 to cpe:2.3:o:cisco:ios:12.2\(1\)
def unquote(cpe):
    return cpe_conversion.unquote(cpe)


# Generates a human readable title from a CPE 2.3 string
def generate_title(cpe):
    title = ""

    cpe_split = cpe.split(":")
    # Do a very basic test to see if the CPE is valid
    if len(cpe_split) == 13:

        # Combine vendor, product and version
        title = " ".join(cpe_split[3:6])

        # If "other" is specified, add it to the title
        if cpe_split[12] != "*":
            title += cpe_split[12]

        # Capitilize each word
        title = title.title()

        # If the target_sw is defined, add "for <target_sw>" to title
        if cpe_split[10] != "*":
            title += " for " + cpe_split[10]

        # In CPE 2.3 spaces are replaced with underscores. Undo it
        title = title.replace("_", " ")

        # Special characters are escaped with \. Undo it
        title = title.replace("\\", "")

    return title



from pprint import pformat

import click

from CveXplore.cli_cmds.cli_utils.utils import printer
from CveXplore.cli_cmds.mutex_options.mutex import Mutex


@click.group(
    "search",
    invoke_without_command=True,
    help="Perform search queries on a single collection",
)
@click.option(
    "-c",
    "--collection",
    required=True,
    type=click.Choice(["capec", "cpe", "cwe", "via4", "cves"], case_sensitive=False),
    help="Collection to query",
)
@click.option("-f", "--field", required=True, help="Field to query")
@click.option("-v", "--value", required=True, help="Value to query")
@click.option("-l", "--limit", default=10, help="Query limit")
@click.option(
    "--pretty",
    is_flag=True,
    help="Pretty print the output",
    cls=Mutex,
    not_required_if=["output"],
)
@click.option(
    "-o",
    "--output",
    default="json",
    help="Set the desired output format",
    type=click.Choice(["json", "csv", "xml", "html"], case_sensitive=False),
    cls=Mutex,
    not_required_if=["pretty"],
)
@click.pass_context
def search_cmd(ctx, collection, field, value, limit, pretty, output):
    ret_list = ctx.obj["data_source"].get_single_store_entries(
        (collection, {field: value.upper()}), limit=limit
    )

    result = [result.to_dict() for result in ret_list]

    if ctx.invoked_subcommand is None:
        printer(input_data=result, pretty=pretty, output=output)
    else:
        if pretty:
            ctx.obj["RESULT"] = pformat(result, indent=4)
        else:
            ctx.obj["RESULT"] = result


@search_cmd.command("less", help="Lets you scroll through the returned results")
@click.pass_context
def less_cmd(ctx):
    click.echo_via_pager(ctx.obj["RESULT"])



import asyncio
import json
import logging
import math
import random
import time
from collections import namedtuple
from datetime import datetime, timedelta
from json import JSONDecodeError
from urllib.parse import urlencode

import aiohttp as aiohttp
import requests
from aiohttp import ContentTypeError
from aioretry import retry, RetryPolicyStrategy, RetryInfo
from requests import Response

from CveXplore.api.api_base_class import ApiBaseClass
from CveXplore.common.config import Configuration
from CveXplore.database.maintenance.LogHandler import UpdateHandler
from CveXplore.errors.apis import (
    ApiErrorException,
    ApiDataError,
    ApiDataRetrievalFailed,
)

logging.setLoggerClass(UpdateHandler)


class NvdNistApi(ApiBaseClass):
    def __init__(
        self,
        baseurl: str = "https://services.nvd.nist.gov",
        api_path: str = "2.0",
        user_agent: str = "CveXplore",
    ):
        super().__init__(baseurl, api_path=api_path, user_agent=user_agent)

        self.config = Configuration()

        self.logger = logging.getLogger("NvdNistApi")

        if self.config.NVD_NIST_API_KEY is not None:
            self.api_key = self.config.NVD_NIST_API_KEY
            self.set_header_field("apiKey", self.api_key)
            self.api_key_limit = False
        else:
            self.api_key_limit = True

        self.filter_rejected = False

        if self.config.NVD_NIST_NO_REJECTED:
            self.filter_rejected = True

        self.datasource = namedtuple("datasource", "CVE CPE")(1, 2)

        self.datasource_mapping = {1: "cves", 2: "cpes"}

        self.max_page_length = namedtuple("max_page_length", "CVE CPE")(2000, 10000)

    def get_url_only(self, resource: dict = None, data: int = 1) -> str:

        return self._build_url(resource=resource, data=data)

    def _build_url(self, resource: dict = None, data: int = 1) -> str:

        if resource is not None:
            resource = urlencode(resource)
            if data == self.datasource.CVE:
                if self.filter_rejected:
                    return f"{self.baseurl}/rest/json/{self.datasource_mapping[data]}/{self.api_path}/?noRejected&{resource}"
                else:
                    return f"{self.baseurl}/rest/json/{self.datasource_mapping[data]}/{self.api_path}/?{resource}"
            else:
                return f"{self.baseurl}/rest/json/{self.datasource_mapping[data]}/{self.api_path}/?{resource}"
        else:
            if data == self.datasource.CVE:
                if self.filter_rejected:
                    return f"{self.baseurl}/rest/json/{self.datasource_mapping[data]}/{self.api_path}/?noRejected"
                else:
                    return f"{self.baseurl}/rest/json/{self.datasource_mapping[data]}/{self.api_path}/"
            else:
                return f"{self.baseurl}/rest/json/{self.datasource_mapping[data]}/{self.api_path}/"

    def _connect(
        self,
        method,
        resource: dict,
        session: requests.Session,
        data=None,
        timeout=60,
        return_response_object=False,
    ):

        requests.packages.urllib3.disable_warnings()

        request_api_resource = {
            "headers": self.myheaders,
            "verify": self.verify,
            "timeout": timeout,
            "proxies": self.proxies,
        }

        try:
            self.logger.debug(f"Sending request: resource={resource}, data={data}")
            r = session.get(
                self._build_url(resource, data=data), **request_api_resource
            )

            try:
                if isinstance(r, Response):
                    if return_response_object:
                        return r
                    if r.status_code >= 400:
                        the_response = json.loads(r.text)
                        raise requests.exceptions.ConnectionError(the_response)
                    else:
                        the_response = json.loads(r.text)
            except JSONDecodeError:
                if r.headers["content-type"] == "text/plain":
                    the_response = r.text
                else:
                    the_response = r

            return the_response

        except requests.exceptions.ConnectionError as err:
            raise requests.exceptions.ConnectionError(err)
        except Exception as err:
            raise Exception(err)

    def __repr__(self):
        """return a string representation of the obj"""
        return f"<< NvdNistApi:{self.baseurl} >>"

    def get_cves_from_start_year(self):

        start_date = datetime(int(self.config.CVE_START_YEAR), 1, 1, 0, 0, 0, 0)
        start_date_iso = start_date.isoformat()

        end_date = start_date + timedelta(days=120)
        end_date_iso = end_date.isoformat()

        resource = {
            "pubStartDate": f"{start_date_iso}",
            "pubEndDate": f"{end_date_iso}",
        }

        return self.call("GET", resource=resource, data=self.datasource.CVE)

    def check_date_range(
        self,
        resource: dict = None,
        last_mod_start_date: datetime = None,
        last_mod_end_date: datetime = None,
    ):

        # Check if diff > 120 days
        diff = last_mod_end_date - last_mod_start_date

        if diff.days > 120:
            delta = diff.days - 120
            last_mod_end_date = last_mod_end_date - timedelta(delta)
            self.logger.warning(
                f"Requested timeframe exceeds the 120 days limit; capping the requested date range to 120 days. "
                f"End date is now: {last_mod_end_date.isoformat()}"
            )

        resource["lastModStartDate"] = last_mod_start_date.isoformat()
        resource["lastModEndDate"] = last_mod_end_date.isoformat()

        return resource

    def get_count(
        self,
        datasource: int = 1,
        last_mod_start_date: datetime = None,
        last_mod_end_date: datetime = None,
    ):
        resource = {"resultsPerPage": 1}

        if last_mod_start_date is not None and last_mod_end_date is not None:
            resource = self.check_date_range(
                resource=resource,
                last_mod_start_date=last_mod_start_date,
                last_mod_end_date=last_mod_end_date,
            )

        ret_data = self.call(self.methods.GET, resource=resource, data=datasource)

        if not isinstance(ret_data, Response):
            return ret_data["totalResults"]
        else:
            raise ApiDataRetrievalFailed

    def get_all_data(
        self,
        data_type: str,
        last_mod_start_date: datetime = None,
        last_mod_end_date: datetime = None,
    ):

        resource = {}

        if last_mod_start_date is not None and last_mod_end_date is not None:
            self.logger.debug(f"Getting all updated {data_type}s....")
            resource = self.check_date_range(
                resource=resource,
                last_mod_start_date=last_mod_start_date,
                last_mod_end_date=last_mod_end_date,
            )
        else:
            self.logger.debug(f"Getting all {data_type}s...")

        data = self.get_count(
            getattr(self.datasource, data_type.upper()),
            last_mod_start_date=last_mod_start_date,
            last_mod_end_date=last_mod_end_date,
        )

        if isinstance(data, int):

            for each_data in ApiData(
                results_per_page=getattr(self.max_page_length, data_type.upper()),
                start_index=0,
                total_results=data,
                api_handle=self,
                data_source=getattr(self.datasource, data_type.upper()),
                resource=resource,
            ):
                yield each_data
        else:
            raise ApiErrorException


class ApiData(object):
    def __init__(
        self,
        results_per_page,
        start_index,
        total_results,
        api_handle: NvdNistApi,
        data_source,
        resource=None,
    ):
        self.results_per_page = results_per_page
        self.start_index = start_index
        self.total_results = total_results
        self.api_handle = api_handle
        self.data_source = data_source
        self.resource = resource

    def __iter__(self):
        return ApiDataIterator(self)


def retry_policy(info: RetryInfo) -> RetryPolicyStrategy:
    """
    - It will always retry until succeeded
    - If fails for the first time, it will retry immediately,
    - If it fails again,
      aioretry will perform a 100ms delay before the second retry,
      200ms delay before the 3rd retry,
      the 4th retry immediately,
      100ms delay before the 5th retry,
      etc...
    """
    max_retries = 5
    backoff_in_ms = 0.2 * 2**info.fails + random.uniform(0, 1) * 4
    logger = logging.getLogger("RetryPolicy")

    if info.fails != max_retries:
        logger.debug(f"Current backoff: {backoff_in_ms}")

        logger.debug(f"Retrying {info.fails + 1}/{max_retries}")

        return False, backoff_in_ms

    else:
        return True, 0


class ApiDataIterator(object):
    def __init__(self, api_data: ApiData):
        self.logger = logging.getLogger("ApiDataIterator")

        self._page_length = api_data.results_per_page
        self._total_results = api_data.total_results
        self._current_index = api_data.start_index
        self.api_data = api_data

        self.sem_factor = 6

        if not self.api_data.api_handle.api_key_limit:
            self.sem_factor = 0.6

        self.logger.debug(f"Using sem factor: {self.sem_factor}")

        self.first_iteration = True

        self.last_stop_time = 0

        self.workload = None

    def __iter__(self):
        return self

    def __next__(self):

        if (
            self._current_index < self._total_results
            or self._page_length == self._total_results
        ):

            start_time = time.time()

            if not self.last_stop_time == 0:
                # adhering to best practices @https://nvd.nist.gov/general/news/API-Key-Announcement
                # which advises to sleep for 6 seconds, so we add this to the 30 seconds rolling window; hence 36
                sleep_time = start_time - self.last_stop_time
                if sleep_time <= 36:
                    self.logger.debug(
                        f"36 second window not expired; sleeping for : {36 - sleep_time}"
                    )
                    time.sleep(36 - sleep_time)

            self.logger.debug(f"Starting download run...")

            self.workload = []

            if self.api_data.api_handle.api_key_limit:
                batch_range = 5
            else:
                batch_range = 45

            for i in range(batch_range):
                if not self.first_iteration:
                    new_start_index = self._current_index + self._page_length
                else:
                    new_start_index = self._current_index
                    self.first_iteration = False

                self._current_index = new_start_index
                resource = {"startIndex": new_start_index}
                if self.api_data.resource is not None:
                    resource = {**resource, **self.api_data.resource}
                if new_start_index > self._total_results:
                    break
                self.workload.append(
                    self.api_data.api_handle.get_url_only(
                        resource=resource, data=self.api_data.data_source
                    )
                )

            ret_data = self.process_async()

            self.last_stop_time = time.time()
            elapsed_time = self.last_stop_time - start_time

            self.logger.debug(f"Elapsed run time: {elapsed_time}")

            return ret_data

        raise StopIteration

    def process_async(self):

        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)
        results = loop.run_until_complete(self.fetch_all(loop))

        return results

    @retry(retry_policy)
    async def fetch(self, session, url):
        try:
            async with session.get(url) as response:
                self.logger.debug(f"Sending request to url: {url}")
                if response.status == 200:
                    data = await response.json()
                    if "format" in data:
                        if data["format"] == "NVD_CPE":
                            if "products" not in data:
                                self.logger.debug(
                                    f"Data received does not contain a products list; raise error to retry!"
                                )
                                raise ApiDataError
                        elif data["format"] == "NVD_CVE":
                            if "vulnerabilities" not in data:
                                self.logger.debug(
                                    f"Data received does not contain a vulnerabilities list; raise error to retry!"
                                )
                                raise ApiDataError
                        else:
                            self.logger.debug(
                                f"Data received does not match expected formatting string; raise error to retry!"
                            )
                            raise ApiDataError
                    else:
                        self.logger.debug(
                            f"Data received does not match expected content; raise error to retry!"
                        )
                        raise ApiDataError
                    return data
                else:
                    if str(response.status).startswith("4"):
                        if "message" in response.headers:
                            self.logger.debug(response.headers["message"])
                    if response.status == 403:
                        self.logger.debug(f"Request forbidden by administrative rules")
                    raise ApiDataRetrievalFailed(url)
        except ApiDataError:
            raise
        except ContentTypeError:
            return ApiDataRetrievalFailed(url)
        finally:
            self.logger.debug(f"Finished request to url: {url}")
            time.sleep(self.sem_factor / 2)

    async def fetch_all(self, loop):
        sem = asyncio.Semaphore(math.ceil(30 / self.sem_factor))
        async with sem:
            async with aiohttp.ClientSession(
                loop=loop,
                headers=self.api_data.api_handle.headers,
                timeout=aiohttp.ClientTimeout(
                    total=30.0, sock_connect=30.0, sock_read=30.0, connect=30.0
                ),
            ) as session:
                results = await asyncio.gather(
                    *[
                        self.fetch(
                            session,
                            entry,
                        )
                        for entry in self.workload
                    ],
                    return_exceptions=True,
                )
                return results



import json
from collections import defaultdict
from pprint import pformat

import click
import pandas as pd
from dicttoxml import dicttoxml


def format_output(format_type, input_list):
    """
    Method to format the input_list based on a given format_type (json|csv|html|xml)

    :param format_type: Format type
    :type format_type: str
    :param input_list: List with dicts that needs formatting
    :type input_list: list
    :return: Formatted output
    :rtype: based on given format_type
    """

    if format_type == "json":
        output = json.dumps(input_list)
    elif format_type == "csv":
        df = pd.DataFrame(input_list)
        output = df.to_csv(index=False)
    elif format_type == "html":
        df = pd.DataFrame(input_list)
        output = df.to_html(index=False)
    elif format_type == "xml":
        if isinstance(input_list, list):
            count = 1
            temp_dict = defaultdict(dict)
            for each in input_list:
                temp_dict["entry_{}".format(count)] = each
                count += 1
            output = dicttoxml(temp_dict, custom_root="CveXplore")
        elif isinstance(input_list, dict):
            output = dicttoxml(input_list, custom_root="CveXplore")

    return output


def printer(input_data, pretty=False, output="json"):
    if isinstance(input_data, list):
        if not pretty:
            click.echo(format_output(output, input_data))
        else:
            click.secho(pformat(input_data, indent=4), bold=True, fg="blue")
    elif isinstance(input_data, dict):
        if not pretty:
            click.echo(input_data)
        else:
            click.secho(pformat(input_data, indent=4), bold=True, fg="blue")



from abc import abstractmethod

from CveXplore.api.nvd_nist.nvd_nist_api import NvdNistApi
from CveXplore.database.maintenance.DownloadHandler import DownloadHandler
from CveXplore.database.maintenance.db_action import DatabaseAction


class NVDApiHandler(DownloadHandler):
    """
    This class handles all JSON related download processing and functions as a base class for specific JSON sources
    processing and downloading
    """

    def __init__(self, feed_type):
        super().__init__(feed_type)

        self.is_update = True

        self.api_handler = NvdNistApi()

    def __repr__(self):
        """return string representation of object"""
        return "<< NVDApiHandler:{} >>".format(self.feed_type)

    def process_item(self, item):
        item = self.process_the_item(item)

        if item is not None:
            if self.is_update:
                return DatabaseAction(
                    action=DatabaseAction.actions.UpdateOne,
                    collection=self.feed_type.lower(),
                    doc=item,
                ).entry
            else:
                return DatabaseAction(
                    action=DatabaseAction.actions.InsertOne,
                    collection=self.feed_type.lower(),
                    doc=item,
                ).entry

    @abstractmethod
    def process_the_item(self, *args):
        raise NotImplementedError

    @abstractmethod
    def file_to_queue(self, *args):
        raise NotImplementedError

    @abstractmethod
    def update(self, **kwargs):
        raise NotImplementedError

    @abstractmethod
    def populate(self, **kwargs):
        raise NotImplementedError



"""
XML Content Handlers
====================
"""
from collections import defaultdict
from xml.sax.handler import ContentHandler


class CapecHandler(ContentHandler):
    """
    Class handling the CAPEC XML parsing
    """

    def __init__(self):
        self.capec = []
        self.Attack_Pattern_Catalog_tag = False
        self.Attack_Patterns_tag = False
        self.Attack_Pattern_tag = False
        self.Attack_step_tag = False
        self.Description_tag = False
        self.Text_tag = False
        self.Prerequisites_tag = False
        self.Prerequisite_tag = False
        self.Mitigations_tag = False
        self.Mitigation_tag = False
        self.Related_Weaknesses_tag = False
        self.Related_Weakness_tag = False
        self.CWE_ID_tag = False
        self.Related_Attack_Patterns = False
        self.Taxonomy_Mappings = False
        self.Taxonomy_Mapping = False
        self.Likelihood_Of_Attack = False
        self.Typical_Severity = False

        self.Execution_Flow = False
        self.Step = True
        self.Phase = True
        self.Attack_Description = True
        self.Technique = True

        self.entry_id = False
        self.entry_name = False

        self.tag = False

        self.id = ""
        self.name = ""

        self.Summary_ch = ""
        self.Prerequisite_ch = ""
        self.Mitigation_ch = ""
        self.CWE_ID_ch = ""
        self.entry_id_ch = ""
        self.entry_name_ch = ""

        self.taxonomy_name = ""
        self.step_name = ""

        self.Step_ch = ""
        self.Phase_ch = ""
        self.Attack_Description_ch = ""
        self.Technique_ch = ""
        self.Likelihood_Of_Attack_ch = ""
        self.Typical_Severity_ch = ""
        self.loa = ""
        self.ts = ""

        self.Summary = []
        self.Prerequisite = []
        self.Solution_or_Mitigation = []
        self.Related_Weakness = []
        self.Related_AttackPatterns = []
        self.techniques = []

        self.taxonomy_mapping = defaultdict(dict)

        self.execution_flow = defaultdict(dict)

    def startElement(self, name, attrs):

        if name == "Attack_Pattern_Catalog":
            self.Attack_Pattern_Catalog_tag = True
        if name == "Attack_Patterns" and self.Attack_Pattern_Catalog_tag:
            self.Attack_Patterns_tag = True
        if name == "Attack_Pattern" and self.Attack_Patterns_tag:
            self.Attack_Pattern_tag = True

        if self.Attack_Pattern_tag:
            self.tag = name
            if self.tag == "Attack_Pattern":
                self.id = attrs.getValue("ID")
                self.name = attrs.getValue("Name")

            if self.tag == "Related_Attack_Patterns":
                self.Related_Attack_Patterns = True

            if self.tag == "Related_Attack_Pattern" and self.Related_Attack_Patterns:
                self.Related_AttackPatterns.append(attrs.get("CAPEC_ID"))

            if self.tag == "Taxonomy_Mappings":
                self.Taxonomy_Mappings = True

            if self.tag == "Taxonomy_Mapping" and self.Taxonomy_Mappings:
                self.Taxonomy_Mapping = True
                self.taxonomy_name = attrs.get("Taxonomy_Name")

            if self.tag == "Entry_ID" and self.Taxonomy_Mappings:
                self.entry_id = True
                self.entry_id_ch = ""

            if self.tag == "Entry_Name" and self.Taxonomy_Mappings:
                self.entry_name = True
                self.entry_name_ch = ""

            if self.tag == "Execution_Flow":
                self.Execution_Flow = True

            if self.tag == "Attack_Step" and self.Execution_Flow:
                self.Attack_step_tag = True

            if self.tag == "Step" and self.Attack_step_tag:
                self.Step = True
                self.Step_ch = ""

            if self.tag == "Phase" and self.Attack_step_tag:
                self.Phase = True
                self.Phase_ch = ""

            if self.tag == "Description" and self.Attack_step_tag:
                self.Attack_Description = True
                self.Attack_Description_ch = ""

            if self.tag == "Technique" and self.Attack_step_tag:
                self.Technique = True
                self.Technique_ch = ""

            if self.tag == "Description" and not self.Attack_step_tag:
                self.Description_tag = True
                self.Summary_ch = ""

            if self.tag == "Prerequisites":
                self.Prerequisites_tag = True
            if name == "Prerequisite" and self.Prerequisites_tag:
                self.Prerequisite_tag = True
                self.Prerequisite_ch = ""

            if self.tag == "Mitigations":
                self.Mitigations_tag = True
            if name == "Mitigation" and self.Mitigations_tag:
                self.Mitigation_tag = True
            if name == "xhtml:p" and self.Mitigation_tag:
                self.Text_tag = True
                self.Mitigation_ch = ""

            if self.tag == "Related_Weaknesses":
                self.Related_Weaknesses_tag = True
            if name == "Related_Weakness" and self.Related_Weaknesses_tag:
                self.Related_Weakness.append(attrs.getValue("CWE_ID"))

            if self.tag == "Likelihood_Of_Attack":
                self.Likelihood_Of_Attack = True
                self.Likelihood_Of_Attack_ch = ""

            if self.tag == "Typical_Severity":
                self.Typical_Severity = True
                self.Typical_Severity_ch = ""

    def characters(self, ch):
        if self.Description_tag:
            self.Summary_ch += ch

        if self.Prerequisite_tag:
            self.Prerequisite_ch += ch

        if self.Text_tag:
            if self.Mitigation_tag:
                self.Mitigation_ch += ch

        if self.entry_id:
            self.entry_id_ch += ch

        if self.entry_name:
            self.entry_name_ch += ch

        if self.Step:
            self.Step_ch += ch

        if self.Phase:
            self.Phase_ch += ch

        if self.Attack_Description:
            self.Attack_Description_ch += ch

        if self.Technique:
            self.Technique_ch += ch

        if self.Likelihood_Of_Attack:
            self.Likelihood_Of_Attack_ch += ch

        if self.Typical_Severity:
            self.Typical_Severity_ch += ch

    def endElement(self, name):
        if name == "Description" and not self.Attack_step_tag:
            self.Summary.append(self.Summary_ch.rstrip())
            if self.Summary_ch != "":
                self.Summary_ch = ""
            self.Description_tag = False

        if name == "Entry_ID":
            self.entry_id = False

        if name == "Entry_Name":
            self.entry_name = False

            entry_id = self.entry_id_ch.rstrip()

            cut_entry = entry_id.split(".")

            url = ""

            if self.taxonomy_name == "ATTACK":
                if len(cut_entry) == 1:
                    # no subtechnique use plain entry_id
                    url = "https://attack.mitre.org/techniques/T{}".format(entry_id)
                else:
                    # attack with subtechniques use cut_entry
                    url = "https://attack.mitre.org/techniques/T{}/{}".format(
                        cut_entry[0], cut_entry[1]
                    )

            elif self.taxonomy_name == "WASC":

                if "/" in self.entry_name_ch:
                    url = "http://projects.webappsec.org/{}".format(
                        self.entry_name_ch.replace("/", " and ").replace(" ", "-")
                    )
                else:
                    url = "http://projects.webappsec.org/{}".format(
                        self.entry_name_ch.replace(" ", "-")
                    )

            elif self.taxonomy_name == "OWASP Attacks":
                entry_id = "Link"

                url = "https://owasp.org/www-community/attacks/{}".format(
                    self.entry_name_ch.replace(" ", "_")
                )

            self.taxonomy_mapping[self.taxonomy_name][
                self.entry_id_ch.rstrip().replace(".", "_")
            ] = {
                "Entry_ID": entry_id,
                "Entry_Name": self.entry_name_ch.rstrip(),
                "URL": url,
            }

            if self.entry_id_ch != "":
                self.entry_id_ch = ""

            if self.entry_name_ch != "":
                self.entry_name_ch = ""

        if name == "Taxonomy_Mappings":
            self.Taxonomy_Mappings = False

        if name == "Taxonomy_Mapping":
            self.Taxonomy_Mapping = False

        if name == "Step":
            self.step_name = self.Step_ch.rstrip()
            self.Step = False

        if name == "Phase":
            self.Phase = False

        if name == "Description" and self.Attack_step_tag:
            self.Attack_Description = False

            self.execution_flow[self.step_name] = {
                "Phase": self.Phase_ch.rstrip(),
                "Description": self.Attack_Description_ch.rstrip(),
                "Techniques": [],
            }

            if self.Step_ch != "":
                self.Step_ch = ""

            if self.Phase_ch != "":
                self.Phase_ch = ""

            if self.Attack_Description_ch != "":
                self.Attack_Description_ch = ""

        if name == "Technique" and self.Attack_step_tag:
            if self.Technique_ch != "":
                self.execution_flow[self.step_name]["Techniques"].append(
                    self.Technique_ch.rstrip()
                )
                self.Technique_ch = ""
            self.Technique = False

        if name == "Attack_Step":
            self.Attack_step_tag = False

        if name == "Execution_Flow":
            self.Execution_Flow = False

        if name == "Prerequisite":
            if self.Prerequisite_ch != "":
                self.Prerequisite.append(self.Prerequisite_ch.rstrip())
            self.Prerequisite_tag = False
        if name == "Mitigation":
            if self.Mitigation_ch != "":
                self.Solution_or_Mitigation.append(self.Mitigation_ch.rstrip())
                self.Mitigation_ch = ""
            self.Mitigation_tag = False

        if name == "Prerequisites":
            self.Prerequisites_tag = False
        if name == "Mitigations":
            self.Mitigations_tag = False
        if name == "Related_Weaknesses":
            self.Related_Weaknesses_tag = False

        if name == "Related_Attack_Patterns":
            self.Related_Attack_Patterns = False

        if name == "Likelihood_Of_Attack":
            self.Likelihood_Of_Attack = False
            self.loa = self.Likelihood_Of_Attack_ch.rstrip()
            self.Likelihood_Of_Attack_ch = ""

        if name == "Typical_Severity":
            self.Typical_Severity = False
            self.ts = self.Typical_Severity_ch.rstrip()
            self.Typical_Severity_ch = ""

        if name == "Attack_Pattern":
            if not self.name.startswith("DEPRECATED"):
                self.capec.append(
                    {
                        "name": self.name,
                        "id": self.id,
                        "summary": "\n".join(self.Summary),
                        "prerequisites": " ".join(self.Prerequisite),
                        "solutions": " ".join(self.Solution_or_Mitigation),
                        "related_capecs": sorted(self.Related_AttackPatterns),
                        "related_weakness": sorted(self.Related_Weakness),
                        "taxonomy": dict(self.taxonomy_mapping),
                        "execution_flow": dict(self.execution_flow),
                        "loa": self.loa,
                        "typical_severity": self.ts,
                    }
                )
            self.Summary = []
            self.Prerequisite = []
            self.Solution_or_Mitigation = []
            self.Related_Weakness = []
            self.Related_AttackPatterns = []
            self.techniques = []

            self.taxonomy_mapping = defaultdict(dict)

            self.execution_flow = defaultdict(dict)

            self.Attack_Pattern_tag = False
        if name == "Attack_Patterns":
            self.Attack_Patterns_tag = False
        if name == "Attack_Pattern_Catalog":
            self.Attack_Pattern_Catalog_tag = False


class CWEHandler(ContentHandler):
    """
    Class handling the CWE XML parsing
    """

    def __init__(self):
        self.cwe = []
        self.description_tag = False
        self.category_tag = False
        self.weakness_tag = False
        self.weakness_relationships_tag = False
        self.category_relationships_tag = False

    def startElement(self, name, attrs):

        if name == "Weakness":
            self.weakness_tag = True
            self.statement = ""
            self.weaknessabs = attrs.get("Abstraction")
            self.name = attrs.get("Name")
            self.idname = attrs.get("ID")
            self.status = attrs.get("Status")
            if not self.name.startswith("DEPRECATED"):
                self.cwe.append(
                    {
                        "name": self.name,
                        "id": self.idname,
                        "status": self.status,
                        "weaknessabs": self.weaknessabs,
                    }
                )

        elif name == "Category":
            self.category_tag = True
            self.category_name = attrs.get("Name")
            self.category_id = attrs.get("ID")
            self.category_status = attrs.get("Status")
            if not self.category_name.startswith("DEPRECATED"):
                self.cwe.append(
                    {
                        "name": self.category_name,
                        "id": self.category_id,
                        "status": self.category_status,
                        "weaknessabs": "Category",
                    }
                )

        elif name == "Description" and self.weakness_tag:
            self.description_tag = True
            self.description = ""

        elif name == "Summary" and self.category_tag:
            self.description_tag = True
            self.description = ""

        elif name == "Relationships" and self.category_tag:
            self.category_relationships_tag = True
            self.relationships = []

        elif name == "Related_Weaknesses" and self.weakness_tag:
            self.weakness_relationships_tag = True
            self.relationships = []

        elif name == "Related_Weakness" and self.weakness_relationships_tag:
            self.relationships.append(attrs.get("CWE_ID"))

        elif name == "Has_Member" and self.category_relationships_tag:
            self.relationships.append(attrs.get("CWE_ID"))

    def characters(self, ch):
        if self.description_tag:
            self.description += ch.replace("       ", "")

    def endElement(self, name):
        if name == "Description" and self.weakness_tag:
            self.description_tag = False
            self.description = self.description + self.description
            self.cwe[-1]["Description"] = self.description.replace("\n", "")
        if name == "Summary" and self.category_tag:
            self.description_tag = False
            self.description = self.description + self.description
            self.cwe[-1]["Description"] = self.description.replace("\n", "")
        elif name == "Weakness" and self.weakness_tag:
            self.weakness_tag = False
        elif name == "Category" and self.category_tag:
            self.category_tag = False

        elif name == "Related_Weaknesses" and self.weakness_tag:
            self.weakness_relationships_tag = False
            self.cwe[-1]["related_weaknesses"] = self.relationships

        elif name == "Relationships" and self.category_tag:
            self.category_relationships_tag = False
            self.cwe[-1]["relationships"] = self.relationships



"""
via4
====
"""
from CveXplore.common.data_source_connection import DatasourceConnection


class Via4(DatasourceConnection):
    """
    Via4 database object
    """

    def __init__(self, **kwargs):
        super().__init__("via4")

        for each in kwargs:
            setattr(self, each, kwargs[each])

    def to_dict(self):
        """
        Method to convert the entire object to a dictionary

        :return: Data from object
        :rtype: dict
        """

        return {k: v for (k, v) in self.__dict__.items() if not k.startswith("_")}

    def __eq__(self, other):
        return self.__dict__ == other.__dict__

    def __ne__(self, other):
        return self.__dict__ != other.__dict__

    def __repr__(self):
        """String representation of object"""
        return "<< Via4:{} >>".format(self.id)



"""
Download Handler
================
"""
import datetime
import gzip
import json
import logging
import os
import sys
import tempfile
import threading
import time
import zipfile
from abc import ABC, abstractmethod
from datetime import timedelta
from io import BytesIO
from itertools import islice
from shutil import copy

import pymongo
import requests
from dateutil.parser import parse as parse_datetime
from pymongo.errors import BulkWriteError, InvalidOperation
from requests.adapters import HTTPAdapter
from tqdm.contrib.concurrent import thread_map
from urllib3 import Retry

from CveXplore.common.config import Configuration
from CveXplore.database.connection.mongo_db import MongoDBConnection
from .LogHandler import UpdateHandler
from .worker_q import WorkerQueue

thread_local = threading.local()
logging.setLoggerClass(UpdateHandler)

logging.getLogger("urllib3").setLevel(logging.WARNING)


class DownloadHandler(ABC):
    """
    DownloadHandler is the base class for all downloads and subsequent processing of the downloaded content.
    Each download script has a derived class which handles specifics for that type of content / download.
    """

    def __init__(self, feed_type, prefix=None):
        self._end = None

        self.feed_type = feed_type

        self.prefix = prefix

        self.queue = WorkerQueue(name=self.feed_type)

        self.file_queue = WorkerQueue(name=f"{self.feed_type}:files")
        self.file_queue.clear()

        self.progress_bar = None

        self.last_modified = None

        self.do_process = True

        database = MongoDBConnection(**json.loads(os.getenv("MONGODB_CON_DETAILS")))

        self.database = database._dbclient

        self.logger = logging.getLogger("DownloadHandler")

        self.config = Configuration()

    def __repr__(self):
        """return string representation of object"""
        return "<< DownloadHandler:{} >>".format(self.feed_type)

    def get_session(
        self,
        retries=3,
        backoff_factor=0.3,
        status_forcelist=(429, 500, 502, 503, 504),
        session=None,
    ):
        """
        Method for returning a session object per every requesting thread
        """

        proxies = {"http": self.config.getProxy(), "https": self.config.getProxy()}

        if not hasattr(thread_local, "session"):
            session = session or requests.Session()
            retry = Retry(
                total=retries,
                read=retries,
                connect=retries,
                backoff_factor=backoff_factor,
                status_forcelist=status_forcelist,
            )

            session.proxies.update(proxies)

            adapter = HTTPAdapter(max_retries=retry)
            session.mount("http://", adapter)
            session.mount("https://", adapter)

            thread_local.session = session

        return thread_local.session

    def process_downloads(self, sites):
        """
        Method to download and process files

        :param sites: List of file to download and process
        :type sites: list
        :return:
        :rtype:
        """

        start_time = time.time()

        thread_map(self.download_site, sites, desc="Downloading files")

        if self.do_process:
            thread_map(
                self.file_to_queue,
                self.file_queue.getall(),
                desc="Processing downloaded files",
            )

            chunks = []

            for batch in iter(lambda: list(islice(self.queue, 10000)), []):
                chunks.append(batch)

            thread_map(
                self._db_bulk_writer, chunks, desc="Transferring queue to database"
            )

            # checking if last-modified was in the response headers and not set to default
            if "01-01-1970" != self.last_modified.strftime("%d-%m-%Y"):
                self.setColUpdate(self.feed_type.lower(), self.last_modified)

        self.logger.info(
            "Duration: {}".format(timedelta(seconds=time.time() - start_time))
        )

    def chunk_list(self, lst, number):
        """
        Yield successive n-sized chunks from lst.

        :param lst: List to be chunked
        :type lst: list
        :param number: Chunk size
        :type number: int
        :return: Chunked list
        :rtype: list
        """
        for i in range(0, len(lst), number):
            yield lst[i : i + number]

    def _db_bulk_writer(self, batch):
        """
        Method to act as worker for writing queued entries into the database

        :param batch: Batch entry
        :type batch: list
        """

        try:
            self.database[self.feed_type.lower()].bulk_write(batch, ordered=False)
        except BulkWriteError as err:
            self.logger.debug(f"Error during bulk write: {err}")
            pass
        except InvalidOperation as err:
            self.logger.debug(f"Got error during bulk update: {err}")
            pass
        except TypeError as err:
            self.logger.debug(f"Error during bulk write: {err}")
            raise

    def store_file(self, response_content, content_type, url):
        """
        Method to store the download based on the headers content type

        :param response_content: Response content
        :type response_content: bytes
        :param content_type: Content type; e.g. 'application/zip'
        :type content_type: str
        :param url: Download url
        :type url: str
        :return: A working directory and a filename
        :rtype: str and str
        """
        wd = tempfile.mkdtemp()
        filename = None

        if (
            content_type == "application/zip"
            or content_type == "application/x-zip"
            or content_type == "application/x-zip-compressed"
            or content_type == "application/zip-compressed"
        ):
            filename = os.path.join(wd, url.split("/")[-1][:-4])
            self.logger.debug("Saving file to: {}".format(filename))

            with zipfile.ZipFile(BytesIO(response_content)) as zip_file:
                zip_file.extractall(wd)

        elif (
            content_type == "application/x-gzip"
            or content_type == "application/gzip"
            or content_type == "application/x-gzip-compressed"
            or content_type == "application/gzip-compressed"
        ):
            filename = os.path.join(wd, url.split("/")[-1][:-3])
            self.logger.debug("Saving file to: {}".format(filename))

            buf = BytesIO(response_content)
            with open(filename, "wb") as f:
                f.write(gzip.GzipFile(fileobj=buf).read())

        elif (
            content_type == "application/json"
            or content_type == "application/xml"
            or content_type == "text/xml"
        ):
            filename = os.path.join(wd, url.split("/")[-1])
            self.logger.debug("Saving file to: {}".format(filename))

            with open(filename, "wb") as output_file:
                output_file.write(response_content)

        elif content_type == "application/local":
            filename = os.path.join(wd, url.split("/")[-1])
            self.logger.debug("Saving file to: {}".format(filename))

            copy(url[7:], filename)

        else:
            self.logger.error(
                "Unhandled Content-Type encountered: {} from url".format(
                    content_type, url
                )
            )
            sys.exit(1)

        return wd, filename

    def download_site(self, url):
        if url[:4] == "file":
            self.logger.info("Scheduling local hosted file: {}".format(url))

            # local file do not get last_modified header; so completely ignoring last_modified check and always asume
            # local file == the last modified file and set to current time.
            self.last_modified = datetime.datetime.now()

            self.logger.debug(
                "Last {} modified value: {} for URL: {}".format(
                    self.feed_type, self.last_modified, url
                )
            )

            wd, filename = self.store_file(
                response_content=b"local", content_type="application/local", url=url
            )

            if filename is not None:
                self.file_queue.put((wd, filename))
            else:
                self.logger.error(
                    "Unable to retrieve a filename; something went wrong when trying to save the file"
                )
                sys.exit(1)

        else:
            self.logger.debug("Downloading from url: {}".format(url))
            session = self.get_session()
            try:
                with session.get(url) as response:
                    try:
                        self.last_modified = parse_datetime(
                            response.headers["last-modified"], ignoretz=True
                        )
                    except KeyError:
                        self.logger.error(
                            "Did not receive last-modified header in the response; setting to default "
                            "(01-01-1970) and force update! Headers received: {}".format(
                                response.headers
                            )
                        )
                        # setting to last_modified to default value
                        self.last_modified = parse_datetime("01-01-1970")

                    self.logger.debug(
                        "Last {} modified value: {} for URL: {}".format(
                            self.feed_type, self.last_modified, url
                        )
                    )

                    i = self.getInfo(self.feed_type.lower())

                    if i is not None:
                        if self.last_modified == i["lastModified"]:
                            self.logger.info(
                                "{}'s are not modified since the last update".format(
                                    self.feed_type
                                )
                            )
                            self.file_queue.getall()
                            self.do_process = False
                    if self.do_process:
                        content_type = response.headers["content-type"]

                        self.logger.debug(
                            "URL: {} fetched Content-Type: {}".format(url, content_type)
                        )

                        wd, filename = self.store_file(
                            response_content=response.content,
                            content_type=content_type,
                            url=url,
                        )

                        if filename is not None:
                            self.file_queue.put((wd, filename))
                        else:
                            self.logger.error(
                                "Unable to retrieve a filename; something went wrong when trying to save the file"
                            )
                            sys.exit(1)
            except Exception as err:
                self.logger.info(
                    "Exception encountered during download from: {}. Please check the logs for more information!".format(
                        url
                    )
                )
                self.logger.error(
                    "Exception encountered during the download from: {}. Error encountered: {}".format(
                        url, err
                    )
                )
                self.do_process = False

    def dropCollection(self, col):
        return self.database[col].drop()

    def getTableNames(self):
        return self.database.list_collection_names()

    def setColInfo(self, collection, field, data):
        self.database[collection].update_one(
            {"db": collection}, {"$set": {field: data}}, upsert=True
        )

    def delColInfo(self, collection):
        self.database["info"].delete_one({"db": collection})

    def getCPEVersionInformation(self, query):
        return self.sanitize(self.database["cpe"].find_one(query))

    def getInfo(self, collection):
        return self.sanitize(self.database["info"].find_one({"db": collection}))

    def sanitize(self, x):
        if type(x) == pymongo.cursor.Cursor:
            x = list(x)
        if type(x) == list:
            for y in x:
                self.sanitize(y)
        if x and "_id" in x:
            x.pop("_id")
        return x

    def setColUpdate(self, collection, date):
        self.database["info"].update_one(
            {"db": collection}, {"$set": {"lastModified": date}}, upsert=True
        )

    @abstractmethod
    def process_item(self, **kwargs):
        raise NotImplementedError

    @abstractmethod
    def file_to_queue(self, *args):
        raise NotImplementedError

    @abstractmethod
    def update(self, **kwargs):
        raise NotImplementedError

    @abstractmethod
    def populate(self, **kwargs):
        raise NotImplementedError



import collections

from pymongo import InsertOne, UpdateOne


class DatabaseAction(object):
    actions = collections.namedtuple("Actions", "InsertOne UpdateOne")(0, 1)

    def __init__(self, action, collection, doc):

        self.action = action
        self.collection = collection
        self.doc = doc

    @property
    def entry(self):
        if self.action == self.actions.InsertOne:
            return InsertOne(self.doc)
        elif self.action == self.actions.UpdateOne:
            return UpdateOne({"id": self.doc["id"]}, {"$set": self.doc}, upsert=True)



"""
Log Handler
===========
"""
import logging
import platform
from logging.config import dictConfig
from logging.handlers import RotatingFileHandler

import colors

from CveXplore.common.config import Configuration


class HostnameFilter(logging.Filter):
    hostname = platform.node()

    def filter(self, record):
        record.hostname = HostnameFilter.hostname
        return True


class HelperLogger(logging.Logger):
    """
    The HelperLogger is used by the application / gui as their logging class and *extends* the default python
    logger.logging class.
    """

    config = Configuration()

    logDict = {
        "version": 1,
        "formatters": {
            "sysLogFormatter": {
                "format": "%(asctime)s - %(name)-8s - %(levelname)-8s - %(message)s"
            },
            "simpleFormatter": {
                "format": "%(asctime)s - %(name)-8s - %(levelname)-8s - %(message)s"
            },
        },
        "handlers": {
            "consoleHandler": {
                "class": "logging.StreamHandler",
                "level": "INFO",
                "stream": "ext://sys.stdout",
                "formatter": "simpleFormatter",
            }
        },
        "root": {"level": "DEBUG", "handlers": ["consoleHandler"]},
    }

    dictConfig(logDict)

    level_map = {
        "debug": "magenta",
        "info": "white",
        "warning": "yellow",
        "error": "red",
        "critical": "red",
    }

    def __init__(self, name, level=logging.NOTSET):
        super().__init__(name, level)

    def debug(self, msg, *args, **kwargs):
        """
        Log ‘msg % args’ with severity ‘DEBUG’ and color *MAGENTA.

        To pass exception information, use the keyword argument exc_info with a true value, e.g.

        .. code-block:: python

            >>> logger.debug(“Houston, we have a thorny problem”)

        :param msg: Message to log
        :type msg: str
        """

        msg = colors.color("{}".format(msg), fg=HelperLogger.level_map["debug"])

        return super(HelperLogger, self).debug(msg, *args, **kwargs)

    def info(self, msg, *args, **kwargs):
        """
        Log ‘msg % args’ with severity ‘INFO’ and color *WHITE*.

        To pass exception information, use the keyword argument exc_info with a true value, e.g.

        .. code-block:: python

            >>> logger.info(“Houston, we have an interesting problem”)

        :param msg: Message to log
        :type msg: str
        """

        msg = colors.color("{}".format(msg), fg=HelperLogger.level_map["info"])

        return super(HelperLogger, self).info(msg, *args, **kwargs)

    def warning(self, msg, *args, **kwargs):
        """
        Log ‘msg % args’ with severity ‘WARNING’ and color *YELLOW*.

        To pass exception information, use the keyword argument exc_info with a true value, e.g.

        .. code-block:: python

            >>> logger.warning(“Houston, we have a bit of a problem”)

        :param msg: Message to log
        :type msg: str
        """

        msg = colors.color("{}".format(msg), fg=HelperLogger.level_map["warning"])

        return super(HelperLogger, self).warning(msg, *args, **kwargs)

    def error(self, msg, *args, **kwargs):
        """
        Log ‘msg % args’ with severity ‘ERROR’ and color *RED*.

        Store logged message to the database for dashboard alerting.

        To pass exception information, use the keyword argument exc_info with a true value, e.g.

        .. code-block:: python

            >>> logger.error(“Houston, we have a major problem”)

        :param msg: Message to log
        :type msg: str
        """

        msg = colors.color("{}".format(msg), fg=HelperLogger.level_map["error"])

        return super(HelperLogger, self).error(msg, *args, **kwargs)

    def critical(self, msg, *args, **kwargs):
        """
        Log ‘msg % args’ with severity ‘CRITICAL’ and color *RED*.

        Store logged message to the database for dashboard alerting.

        To pass exception information, use the keyword argument exc_info with a true value, e.g.

        .. code-block:: python

            >>> logger.critical(“Houston, we have a hell of a problem”)

        :param msg: Message to log
        :type msg: str
        """

        msg = colors.color("{}".format(msg), fg=HelperLogger.level_map["critical"])

        return super(HelperLogger, self).critical(msg, *args, **kwargs)


class UpdateHandler(HelperLogger):
    """
    The UpdateHandler is used by the update process to provide written and visual feedback to the initiator of database
    management tasks.
    """

    def __init__(self, name, level=logging.NOTSET):
        super().__init__(name, level)

        formatter = logging.Formatter(
            "%(asctime)s - %(name)-8s - %(levelname)-8s - %(message)s"
        )

        if self.config.LOGGING_TO_FILE:
            crf = RotatingFileHandler(
                filename=self.config.getUpdateLogFile(),
                maxBytes=self.config.getMaxLogSize(),
                backupCount=self.config.getBacklog(),
            )
            crf.setLevel(logging.DEBUG)
            crf.setFormatter(formatter)
            self.addHandler(crf)



import json
import logging
import os

from CveXplore.database.connection.mongo_db import MongoDBConnection

runPath = os.path.dirname(os.path.realpath(__file__))


class SchemaChecker(object):
    def __init__(self):
        with open(os.path.join(runPath, "../../.schema_version")) as f:
            self.schema_version = json.loads(f.read())

        database = MongoDBConnection(**json.loads(os.getenv("MONGODB_CON_DETAILS")))

        self.dbh = database._dbclient["schema"]

        self.logger = logging.getLogger("SchemaChecker")

    def create_indexes(self):
        # hack for db_updater.py to put this class in the posts variable and run the update method
        self.logger.info("Updating schema version")
        self.update()
        self.logger.info("Update schema version done!")

    def update(self):
        try:

            current_record = list(self.dbh.find({}))

            if len(current_record) != 0:
                current_record[0]["version"] = self.schema_version["version"]
                current_record[0]["rebuild_needed"] = self.schema_version[
                    "rebuild_needed"
                ]

                self.dbh.update_one(
                    {"_id": current_record[0]["_id"]}, {"$set": current_record[0]}
                )
            else:
                current_record = {
                    "version": self.schema_version["version"],
                    "rebuild_needed": self.schema_version["rebuild_needed"],
                }

                self.dbh.insert_one(current_record)
        except AttributeError:
            current_record = {
                "version": self.schema_version["version"],
                "rebuild_needed": self.schema_version["rebuild_needed"],
            }
            self.dbh.insert_one(current_record)



"""
File Handlers
=============
"""
import shutil
from abc import abstractmethod

from CveXplore.database.maintenance.DownloadHandler import DownloadHandler
from CveXplore.database.maintenance.IJSONHandler import IJSONHandler
from CveXplore.database.maintenance.db_action import DatabaseAction


class JSONFileHandler(DownloadHandler):
    """
    This class handles all JSON related download processing and functions as a base class for specific JSON sources
    processing and downloading
    """

    def __init__(self, feed_type, prefix):
        super().__init__(feed_type)

        self.is_update = True

        self.prefix = prefix

        self.ijson_handler = IJSONHandler()

    def __repr__(self):
        """return string representation of object"""
        return "<< JSONFileHandler:{} >>".format(self.feed_type)

    def file_to_queue(self, file_tuple):
        """
        Method responsible for transferring file contents to the worker queue for further processing and inserting them
        into the database

        :param file_tuple: A tuple containing the working dir for processing and the filename of the file to process
        :type file_tuple: tuple
        """

        working_dir, filename = file_tuple

        # adjust the interval counter for debug logging when updating
        if self.is_update:
            interval = 500
        else:
            interval = 5000

        x = 0
        self.logger.debug("Starting processing of file: {}".format(filename))
        for cpe in self.ijson_handler.fetch(filename=filename, prefix=self.prefix):
            self.process_item(item=cpe)
            x += 1
            if x % interval == 0:
                self.logger.debug(
                    "Processed {} entries from file: {}".format(x, filename)
                )

        try:
            self.logger.debug("Removing working dir: {}".format(working_dir))
            shutil.rmtree(working_dir)
        except Exception as err:
            self.logger.error(
                "Failed to remove working dir; error produced: {}".format(err)
            )

    @abstractmethod
    def update(self, **kwargs):
        raise NotImplementedError

    @abstractmethod
    def populate(self, **kwargs):
        raise NotImplementedError


class XMLFileHandler(DownloadHandler):
    """
    This class handles all XML related download processing and functions as a base class for specific XML sources
    processing and downloading
    """

    def __init__(self, feed_type):
        super().__init__(feed_type)
        self.is_update = True

    def __repr__(self):
        """return string representation of object"""
        return "<< XMLFileHandler:{} >>".format(self.feed_type)

    def process_item(self, item):
        """
        Method responsible for putting items into the worker queue as database actions

        :param item: Item processed from XML source and parsed into a dictionary
        :type item: dict
        """

        if self.is_update:
            self.queue.put(
                DatabaseAction(
                    action=DatabaseAction.actions.UpdateOne,
                    collection=self.feed_type.lower(),
                    doc=item,
                )
            )
        else:
            self.queue.put(
                DatabaseAction(
                    action=DatabaseAction.actions.InsertOne,
                    collection=self.feed_type.lower(),
                    doc=item,
                )
            )

    @abstractmethod
    def file_to_queue(self, *args):
        raise NotImplementedError

    @abstractmethod
    def update(self, **kwargs):
        raise NotImplementedError

    @abstractmethod
    def populate(self, **kwargs):
        raise NotImplementedError



"""
Mongodb specific
================
"""

from pymongo.collection import Collection
from pymongo.cursor import Cursor


class CveSearchCursor(Cursor):
    """
    The CveSearchCursor is a custom cursor based on the pymongo cursor which will return database objects instead of
    the raw data from the mongodb database.
    """

    def __init__(self, collection, *args, **kwargs):
        """
        Create a new cve-search cursor.

        :param collection: Reference to a CveSearchCollection object
        :type collection: CveSearchCollection
        """
        super().__init__(collection, *args, **kwargs)

        from CveXplore.objects.capec import Capec
        from CveXplore.objects.cpe import Cpe
        from CveXplore.objects.cves import Cves
        from CveXplore.objects.cwe import Cwe
        from CveXplore.objects.via4 import Via4

        self.database_objects_mapping = {
            "capec": Capec,
            "cpe": Cpe,
            "cwe": Cwe,
            "via4": Via4,
            "cves": Cves,
        }

    @property
    def __empty(self):
        return self._Cursor__empty

    @property
    def __data(self):
        return self._Cursor__data

    @property
    def __manipulate(self):
        return self._Cursor__manipulate

    @property
    def __collection(self):
        return self._Cursor__collection

    @property
    def __collname(self):
        return self._Cursor__collname

    def next(self):
        """
        Advance the cursor and return CveXplore objects
        """
        if self.__empty:
            raise StopIteration
        if len(self.__data) or self._refresh():
            try:
                return self.database_objects_mapping[self.__collname](
                    **self.__data.popleft()
                )
            except KeyError:
                return self.__data.popleft()
        else:
            raise StopIteration

    __next__ = next

    def __repr__(self):
        """Return string representation of this class"""
        return "<< CveSearchCursor:{} >>".format(self.collection)


class CveSearchCollection(Collection):
    """
    The CveSearchCollection is a custom Collection based on the pymongo Collection class which has been altered to
    return a CveSearchCursor reference on the find method.
    """

    def __init__(self, database, name, **kwargs):
        """
        Get / create a custon cve-search Mongo collection.

        :param database: the database to get a collection from
        :type database: MongoDBConnection
        :param name: the name of the collection to get
        :type name: str
        :param kwargs: additional keyword arguments will be passed as options for the create collection command
        :type kwargs: kwargs
        """

        super().__init__(database, name, **kwargs)

    def __repr__(self):
        """Return string representation of this class"""
        return "<< CveSearchCollection:{} >>".format(self.name)

    def find(self, *args, **kwargs):
        """
        Query the database as you would do so with a pymongo Collection.

        :return: Reference to the CveSearchCursor
        :rtype: CveSearchCursor
        """
        return CveSearchCursor(self, *args, **kwargs)



"""
Main Updater
============
"""
import logging
import time

from CveXplore.database.maintenance.DatabaseSchemaChecker import SchemaChecker
from CveXplore.database.maintenance.LogHandler import UpdateHandler
from CveXplore.database.maintenance.Sources_process import (
    CPEDownloads,
    CVEDownloads,
    CWEDownloads,
    CAPECDownloads,
    VIADownloads,
    DatabaseIndexer,
)
from CveXplore.errors import UpdateSourceNotFound

logging.setLoggerClass(UpdateHandler)


class MainUpdater(object):
    """
    The MainUpdater class is the main class for performing database maintenaince tasks
    """

    def __init__(self, datasource):
        """
        Init a new MainUpdater class

        :param datasource: Datasource to update
        :type datasource: MongoDBConnection
        """

        self.datasource = datasource

        self.sources = [
            {"name": "cpe", "updater": CPEDownloads},
            {"name": "cve", "updater": CVEDownloads},
            {"name": "cwe", "updater": CWEDownloads},
            {"name": "capec", "updater": CAPECDownloads},
            {"name": "via4", "updater": VIADownloads},
        ]

        self.posts = [
            {"name": "ensureindex", "updater": DatabaseIndexer},
            {"name": "schema", "updater": SchemaChecker},
        ]

        self.logger = logging.getLogger("MainUpdater")

    def update(self, update_source: str | list = None):
        """
        Method used for updating the database
        """
        if not isinstance(update_source, str | list):
            raise ValueError
        try:
            if update_source is None:

                for source in self.sources:
                    up = source["updater"]()
                    up.update()

            elif isinstance(update_source, list):
                for source in update_source:
                    try:
                        update_this_source = [
                            x for x in self.sources if x["name"] == source
                        ][0]
                        up = update_this_source["updater"]()
                        up.update()
                    except IndexError:
                        raise UpdateSourceNotFound(
                            f"Provided source: {source} could not be found...."
                        )
            else:
                # single string then....
                try:
                    update_this_source = [
                        x for x in self.sources if x["name"] == update_source
                    ][0]
                    up = update_this_source["updater"]()
                    up.update()
                except IndexError:
                    raise UpdateSourceNotFound(
                        f"Provided source: {update_source} could not be found...."
                    )
        except UpdateSourceNotFound:
            raise
        else:
            for post in self.posts:
                indexer = post["updater"]()
                indexer.create_indexes()

        self.datasource.set_handlers_for_collections()

        self.logger.info(f"Database update / initialization complete!")

    def populate(self, populate_source: str | list = None):
        """
        Method used for updating the database
        """
        if not isinstance(populate_source, str | list):
            raise ValueError
        try:
            if populate_source is None:
                for source in self.sources:
                    up = source["updater"]()
                    up.populate()

            elif isinstance(populate_source, list):
                for source in populate_source:
                    try:
                        update_this_source = [
                            x for x in self.sources if x["name"] == source
                        ][0]
                        up = update_this_source["updater"]()
                        up.populate()
                    except IndexError:
                        raise UpdateSourceNotFound(
                            f"Provided source: {source} could not be found...."
                        )
            else:
                # single string then....
                try:
                    update_this_source = [
                        x for x in self.sources if x["name"] == populate_source
                    ][0]
                    up = update_this_source["updater"]()
                    up.populate()
                except IndexError:
                    raise UpdateSourceNotFound(
                        f"Provided source: {populate_source} could not be found...."
                    )
        except UpdateSourceNotFound:
            raise
        else:
            for post in self.posts:
                indexer = post["updater"]()
                indexer.create_indexes()

        self.datasource.set_handlers_for_collections()

        self.logger.info(f"Database update / initialization complete!")

    def initialize(self):
        """
        Method to initialize a new (fresh) instance of a cvedb database
        """

        cpe_pop = CPEDownloads()
        cpe_pop.populate()

        self.logger.info(
            f"Sleeping for 30 seconds between CPE and CVE database population.."
        )
        time.sleep(30)

        cve_pop = CVEDownloads()
        cve_pop.populate()

        self.update()



import click

from CveXplore.cli_cmds.cve_cmds import commands as group2
from CveXplore.cli_cmds.db_cmds import commands as group4
from CveXplore.cli_cmds.search_cmds import commands as group1
from CveXplore.cli_cmds.stats_cmds import commands as group3
from CveXplore.main import CveXplore


@click.group(invoke_without_command=True)
@click.option("-v", "--version", is_flag=True, help="Show the current version and exit")
@click.pass_context
def main(ctx, version):
    ctx.obj = {"data_source": CveXplore()}
    if version:
        click.echo(ctx.obj["data_source"].version)
        exit(0)


main.add_command(group1.search_cmd)
main.add_command(group2.cve_cmd)
main.add_command(group3.stats_cmd)
main.add_command(group4.db_cmd)



"""
Data source connection
======================
"""
import json
import os

from CveXplore.api.connection.api_db import ApiDatabaseSource
from CveXplore.database.connection.mongo_db import MongoDBConnection


class DatasourceConnection(object):
    """
    The DatasourceConnection class handles the connection to the data source and is the base class for the database
    objects and generic database functions
    """

    # hack for documentation building
    if json.loads(os.getenv("DOC_BUILD"))["DOC_BUILD"] != "YES":
        __DATA_SOURCE_CONNECTION = (
            ApiDatabaseSource(**json.loads(os.getenv("API_CON_DETAILS")))
            if os.getenv("API_CON_DETAILS")
            else MongoDBConnection(**json.loads(os.getenv("MONGODB_CON_DETAILS")))
        )

    def __init__(self, collection):
        """
        Create a DatasourceConnection object

        :param collection: Define the collection to connect to
        :type collection: str
        """
        self.__collection = collection

    @property
    def _datasource_connection(self):
        return DatasourceConnection.__DATA_SOURCE_CONNECTION

    @property
    def _datasource_collection_connection(self):
        return getattr(
            DatasourceConnection.__DATA_SOURCE_CONNECTION,
            "store_{}".format(self.__collection),
        )

    @property
    def _collection(self):
        return self.__collection



import json
import os

import click

from CveXplore.cli_cmds.cli_utils.utils import printer
from CveXplore.common.config import Configuration


@click.group(
    "database", invoke_without_command=True, help="Database update / populate commands"
)
@click.pass_context
def db_cmd(ctx):
    pass


@db_cmd.group("update", invoke_without_command=True, help="Update the database")
@click.pass_context
def update_cmd(ctx):
    ctx.obj["data_source"].database.update()


@db_cmd.group("initialize", invoke_without_command=True, help="Initialize the database")
@click.pass_context
def initialize_cmd(ctx):
    ctx.obj["data_source"].database.initialize()


@db_cmd.group("sources", invoke_without_command=True, help="Database source management")
@click.pass_context
def sources_cmd(ctx):
    pass


@sources_cmd.group("show", invoke_without_command=True, help="Show sources")
@click.option("--pretty", is_flag=True, help="Pretty print the output")
@click.pass_context
def show_cmd(ctx, pretty):
    config = Configuration()

    if ctx.invoked_subcommand is None:
        printer(input_data=config.SOURCES, pretty=pretty)
    else:
        printer(input_data=config.SOURCES, pretty=pretty)


@sources_cmd.group("set", invoke_without_command=True, help="Set sources")
@click.option(
    "-k",
    "--key",
    help="Set the source key",
    type=click.Choice(["capec", "cpe", "cwe", "via4", "cves"], case_sensitive=False),
)
@click.option("-v", "--value", help="Set the source key value")
@click.pass_context
def set_cmd(ctx, key, value):
    config = Configuration()

    sources = config.SOURCES

    sources[key] = value

    with open(os.path.join(config.USER_HOME_DIR, ".sources.ini"), "w") as f:
        f.write(json.dumps(sources))

    printer(input_data={"SOURCES SET TO": sources}, pretty=True)


@sources_cmd.group("reset", invoke_without_command=True, help="Set sources")
@click.pass_context
def reset_cmd(ctx):
    config = Configuration()

    sources = config.DEFAULT_SOURCES

    with open(os.path.join(config.USER_HOME_DIR, ".sources.ini"), "w") as f:
        f.write(json.dumps(sources))

    printer(input_data={"SOURCES RESET TO": sources}, pretty=True)



"""
cwe
===
"""
from CveXplore.common.data_source_connection import DatasourceConnection


class Cwe(DatasourceConnection):
    """
    Cwe database object
    """

    def __init__(self, **kwargs):

        super().__init__("cwe")

        for each in kwargs:
            setattr(self, each, kwargs[each])

    def iter_related_weaknessess(self):
        """
        Generator function for iterating the related weaknesses from the current weakness object

        :return: Cwe object
        :rtype: Cwe
        """
        if hasattr(self, "related_weaknesses"):
            if len(self.related_weaknesses) != 0:
                for each in self.related_weaknesses:
                    cwe_doc = self._datasource_connection.store_cwe.find_one(
                        {"id": each}
                    )

                    yield cwe_doc

    def iter_related_capecs(self):
        """
         Generator function for iterating the related capecs from the current weakness object

        :return: Capec object
        :rtype: Capec
        """

        related_capecs = self._datasource_connection.store_capec.find(
            {"related_weakness": self.id}
        )

        for each in related_capecs:
            yield each

    def to_dict(self):
        """
        Method to convert the entire object to a dictionary

        :return: Data from object
        :rtype: dict
        """

        return {k: v for (k, v) in self.__dict__.items() if not k.startswith("_")}

    def __eq__(self, other):
        return self.__dict__ == other.__dict__

    def __ne__(self, other):
        return self.__dict__ != other.__dict__

    def __repr__(self):
        """String representation of object"""
        return "<< Cwe:{} >>".format(self.id)



from queue import Empty, Queue

from CveXplore.database.maintenance.db_action import DatabaseAction


class WorkerQueue(Queue):
    def __init__(self, name, maxsize=0):

        super().__init__(maxsize)
        self.name = name
        self.maxsize = maxsize

        self.closed = False

    def __len__(self):
        self.qsize()

    def __repr__(self):
        return "<< WorkerQueue:{} >>".format(self.name)

    def __iter__(self):
        return self

    def __next__(self):
        try:
            item = self.get(timeout=1)
            if item is not None:
                if isinstance(item, DatabaseAction):
                    item = item.entry
                return item
            else:
                raise StopIteration
        except Empty:
            raise StopIteration

    def getall(self):
        return list(iter(self))

    def clear(self):
        with self.not_empty:
            self.queue.clear()
            self.not_full.notify()
